# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "12/01/2022"


from contextlib import AbstractContextManager
from typing import Optional, Iterable
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.gui.control.datalist import TomoObjList
from tomwer.gui.visualization.tomoobjoverview import TomoObjOverview
from tomwer.core.tomwer_object import TomwerObject
from silx.gui import qt
from silx.gui.utils import blockSignals
from tomoscan.serie import Serie
from silx.utils.enum import Enum as _Enum
import logging

from tomwer.gui.qfolderdialog import QDataDialog

_logger = logging.getLogger(__name__)


class SerieWidgetDialog(qt.QDialog):
    sigSerieSelected = qt.Signal(Serie)
    """
    emit when a serie is selected / triggered by the user
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.setLayout(qt.QVBoxLayout())
        # add list
        self._widget = SerieWidget()
        self.layout().addWidget(self._widget)
        # add buttons
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._selectButton = qt.QPushButton("Select (active) serie", parent=self)
        self._buttons.addButton(self._selectButton, qt.QDialogButtonBox.ActionRole)
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._selectButton.released.connect(self._serieSelected)

    def getSelectedSerie(self) -> Optional[Serie]:
        return self._widget.getSelectedSerie()

    def _serieSelected(self, *args, **kwargs):
        serie = self.getSelectedSerie()
        if serie is not None:
            self.sigSerieSelected.emit(serie)

    # expose API
    def add(self, tomo_obj):
        self._widget.add(tomo_obj=tomo_obj)


class SerieWidget(qt.QTabWidget):
    sigCurrentSerieChanged = qt.Signal()
    """signal emit when the current serie changes"""

    sigHistoryChanged = qt.Signal()
    """signal emit when the history changed (a serie has been added or removed"""

    sigSerieSend = qt.Signal(Serie)
    """Signal emited when a serie has been send"""

    _HISTORY_MODE = "history"
    _DEFINITION_MODE = "serie definition"

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setWindowTitle("serie of scans")
        self._serieDefinitionWidget = SerieDefinition(parent=self)
        self.addTab(self._serieDefinitionWidget, self._DEFINITION_MODE)
        self._historyWidget = SerieHistoryDialog(parent=self)
        self._historyWidget.setWindowFlags(qt.Qt.Widget)
        self.addTab(self._historyWidget, self._HISTORY_MODE)

        # conenct signal / slot
        self._historyWidget.sigEditSerie.connect(self._serieEditionRequested)
        self._historyWidget.sigSerieSend.connect(self._repeatSerieSend)
        self._historyWidget.sigHistoryUpdated.connect(self._repeatHistoryUpdated)
        self._serieDefinitionWidget.sigSerieChanged.connect(self._repeatSerieChanged)
        self._serieDefinitionWidget.sigSerieSend.connect(self._repeatSerieSend)
        self._serieDefinitionWidget.sigSerieSend.connect(self._historyWidget.addSerie)

    def getHistoryWidget(self):
        return self._historyWidget

    def getDefinitionWidget(self):
        return self._serieDefinitionWidget

    def getSelectedSerie(self) -> Optional[Serie]:
        return self._serieDefinitionWidget.getSelectedSerie()

    def setMode(self, mode: str, definition_mode: Optional[str] = None):
        valid_modes = (self._HISTORY_MODE, self._DEFINITION_MODE)
        if mode == self._HISTORY_MODE:
            self.setCurrentWidget(self._historyWidget)
        elif mode == self._DEFINITION_MODE:
            self.setCurrentWidget(self._serieDefinitionWidget)
            self._serieDefinitionWidget.setMode(definition_mode)
        else:
            raise ValueError(
                f"mode {mode} is no recognized. Valid modes are {valid_modes}"
            )

    def _serieEditionRequested(self, serie: Serie):
        if not isinstance(serie, Serie):
            raise TypeError(f"serie is expected to be a serie not {type(serie)}")
        self.setMode("serie definition", "manual")
        self.getDefinitionWidget().getManualDefinitionWidget().setSerie(serie)

    def _repeatSerieChanged(self, *args, **kwargs):
        self.sigCurrentSerieChanged.emit()

    def _repeatSerieSend(self, serie: Serie):
        self.sigSerieSend.emit(serie)

    def _repeatHistoryUpdated(self, *args, **kwargs):
        self.sigHistoryChanged.emit()

    def add(self, tomo_obj):
        return self._serieDefinitionWidget.addTomoObj(tomo_obj)


class _SerieDefinitionMode(_Enum):
    MANUAL = "manual"
    AUTO = "auto"


class SerieDefinition(qt.QWidget):
    sigSerieChanged = qt.Signal()
    ### signal emit when a the serie defined manually changed

    sigSerieSend = qt.Signal(Serie)
    ### signal emit when a serie is send

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())

        self._modeLabel = qt.QLabel("Mode", self)
        self.layout().addWidget(self._modeLabel, 0, 0, 1, 1)

        self._modeCB = qt.QComboBox(self)
        for mode in _SerieDefinitionMode.values():
            self._modeCB.addItem(mode)
        self.layout().addWidget(self._modeCB, 0, 1, 1, 1)

        self._manualDefWidget = SerieManualFromTomoObj(parent=self)
        self.layout().addWidget(self._manualDefWidget, 1, 0, 1, 2)
        self._manualDefWidget.setWindowFlags(qt.Qt.Widget)

        self._automaticDefWidget = SerieAutomaticDefinitionWidget(parent=self)
        self.layout().addWidget(self._automaticDefWidget, 2, 0, 1, 2)

        # connect signal / slot
        self._modeCB.currentIndexChanged.connect(self._updateVisibility)
        self._manualDefWidget._newSerieWidget.sigUpdated.connect(
            self._repeatSerieChanged
        )

        # set up
        self._updateVisibility()

    def getSelectedSerie(self) -> Optional[Serie]:
        if self.getMode() == _SerieDefinitionMode.MANUAL:
            return self._manualDefWidget.getSerie()
        else:
            raise ValueError(f"mode {self.getMode()} is not handled yet")

    def getMode(self) -> str:
        return _SerieDefinitionMode.from_value(self._modeCB.currentText())

    def setMode(self, mode: str):
        mode = _SerieDefinitionMode.from_value(mode)
        idx = self._modeCB.findText(mode.value)
        self._modeCB.setCurrentIndex(idx)

    def _updateVisibility(self):
        self._manualDefWidget.setVisible(self.getMode() == _SerieDefinitionMode.MANUAL)
        self._automaticDefWidget.setVisible(self.getMode() == _SerieDefinitionMode.AUTO)

    def getManualDefinitionWidget(self):
        return self._manualDefWidget

    def getAutoDefinitionWidget(self):
        return self._automaticDefWidget

    def _repeatSerieChanged(self, *args, **kwargs):
        self.sigSerieChanged.emit()

    def _repeatSerieSend(self, serie: Serie):
        self.sigSerieSend.emit(serie)

    def createManualSerie(self):
        self._manualDefWidget.createSerie()

    def addTomoObj(self, tomo_obj: TomwerObject):
        self._manualDefWidget.addTomoObj(tomo_obj=tomo_obj)

    def setSerieName(self, name: str):
        self._manualDefWidget.setSerieName(name=name)


class _SerieDefinitionTree(qt.QWidget):
    """
    Tree used to define manually serie of scan.
    Drag and drop of files is handled
    """

    sigUpdated = qt.Signal()
    """Signal emit when the serie is updated"""

    class SignalBlocker(AbstractContextManager):
        """Simple context manager to hide / show button dialogs"""

        def __init__(self, serie_definition_widget) -> None:
            super().__init__()
            self.serie_definition_widget = serie_definition_widget

        def __enter__(self):
            self.old_widget = self.serie_definition_widget.blockSignals(True)
            self.old_tree = self.serie_definition_widget._tree.blockSignals(True)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.serie_definition_widget.blockSignals(self.old_widget)
            self.serie_definition_widget._tree.blockSignals(self.old_tree)

    def __init__(self, parent=None, serie_name="my_serie") -> None:
        self._tomo_objs = {}
        # associated serie name (key) to tuple (serie, QTreeWidgetItem)
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._tree = qt.QTreeWidget(self)
        self._tree.setSelectionMode(qt.QAbstractItemView.MultiSelection)
        self._tree.setColumnCount(2)
        self._tree.setHeaderLabels(("serie", "scan ids"))
        self._tree.setItemsExpandable(False)
        self.layout().addWidget(self._tree)

        # set up the tree with the serie name that will stay during the entire
        # life time of the tree
        self._serieItem = qt.QTreeWidgetItem(self._tree)
        self._serieItem.setFlags(self._serieItem.flags() | qt.Qt.ItemIsEditable)
        self._serieItem.setExpanded(True)

        self.setAcceptDrops(True)

        # connect signal / slot
        self._tree.itemChanged.connect(self._updated)

        # set up
        self.setSerieName(name=serie_name)

        # expose API
        self.itemChanged = self._tree.itemChanged

    @property
    def rootItem(self):
        return self._serieItem

    def setSerieName(self, name: str):
        with self.SignalBlocker(self):
            self._serieItem.setText(0, name)
        self.sigUpdated.emit()

    def getSerieName(self):
        return self._serieItem.text(0)

    def addTomoObj(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"{tomo_obj} is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )
        identifier = tomo_obj.get_identifier().to_str()
        if identifier in self._tomo_objs:
            _logger.warning(f"scan {identifier} already part of the serie")
            return

        with self.SignalBlocker(self):
            tomo_obj_item = qt.QTreeWidgetItem(self.rootItem)
            tomo_obj_item.setText(1, identifier)
            tomo_obj_item.setFlags(tomo_obj_item.flags() | qt.Qt.ItemIsUserCheckable)
            self._tomo_objs[identifier] = (tomo_obj, tomo_obj_item)
        self.sigUpdated.emit()

    def removeTomoObj(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"{tomo_obj} is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )

        with self.SignalBlocker(self):
            identifier = tomo_obj.get_identifier().to_str()
            if identifier not in self._tomo_objs:
                _logger.warning(f"{identifier} is not in the serie")
            else:
                _, tomo_obj_item = self._tomo_objs.pop(identifier)
                root = self._tree.invisibleRootItem()
                root.removeChild(tomo_obj_item)
        self.sigUpdated.emit()

    @property
    def n_tomo_objs(self):
        return len(self._tomo_objs)

    def setSerie(self, serie: Serie) -> None:
        if not isinstance(serie, Serie):
            raise TypeError(
                f"serie is expected to be an instance of {Serie} not {type(serie)}"
            )

        with self.SignalBlocker(self):
            self.clearTomoObjs()
            self.setSerieName(serie.name)
            for tomo_obj in serie:
                if isinstance(tomo_obj, str):
                    try:
                        tomo_obj = ScanFactory.create_tomo_object_from_identifier(
                            identifier=tomo_obj
                        )
                    except Exception:
                        try:
                            tomo_obj = VolumeFactory.create_tomo_object_from_identifier(
                                identifier=tomo_obj
                            )
                        except Exception:
                            _logger.warning(f"Fail to recreate scan from {tomo_obj}.")
                            return
                elif not isinstance(tomo_obj, TomwerObject):
                    raise TypeError(
                        f"tomo_obj is expected to be an instance of {TomwerObject}. Not {type(tomo_obj)}"
                    )
                self.addTomoObj(tomo_obj)
        self.sigUpdated.emit()

    def getSerie(self, use_identifiers=False) -> Serie:
        scans = [scan for scan, _ in self._tomo_objs.values()]
        return Serie(
            name=self.getSerieName(),
            iterable=scans,
            use_identifiers=use_identifiers,
        )

    def clearTomoObjs(self):
        with self.SignalBlocker(self):
            keys = list(self._tomo_objs.keys())
            for key in keys:
                _, scan_item = self._tomo_objs.pop(key)
                root = self._tree.invisibleRootItem()
                root.removeChild(scan_item)
        self.sigUpdated.emit()

    def setSelectedTomoObjs(self, objs):
        self.clearSelection()
        for scan in objs:
            scan_item = self._getTomoObjItem(scan)
            if scan_item is not None:
                scan_item.setSelected(True)

    def _getTomoObjItem(self, tomo_obj: TomwerObject) -> Optional[qt.QTreeWidgetItem]:
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )
        return self._tomo_objs.get(tomo_obj.get_identifier().to_str(), (None, None))[1]

    def getSelectedTomoObjs(self) -> tuple():
        """return selected scans"""
        selected = []
        for _, (scan, item) in self._tomo_objs.items():
            if item.isSelected():
                selected.append(scan)
        return tuple(selected)

    def removeSelectedTomoObjs(self) -> None:
        with self.SignalBlocker(self):
            for tomo_obj in self.getSelectedTomoObjs():
                self.removeTomoObj(tomo_obj)

    def _updated(self, *args, **kwargs):
        self.sigUpdated.emit()

    def clearSelection(self) -> None:
        self._tree.selectionModel().clearSelection()

    def addScanFromNxFile(self, file_: str, entry: Optional[str] = None):
        try:
            if entry is None:
                scans = ScanFactory.create_scan_objects(scan_path=file_)
            else:
                scans = [ScanFactory.create_scan_object(scan_path=file_, entry=entry)]
        except Exception as e:
            _logger.error(
                "cannot create scan instances from {}. Error is {}".format(file_, e)
            )
        else:
            changed = False
            with self.SignalBlocker(self):
                for scan in scans:
                    if scan is not None:
                        try:
                            self.addTomoObj(tomo_obj=scan)
                        except TypeError:
                            _logger.error(
                                f"fail to add scan {scan}. Invalid type encountered ({type(scan)})"
                            )
                        else:
                            changed = True
            if changed:
                self.sigUpdated.emit()

    def dropEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            for url in event.mimeData().urls():
                self.addScanFromNxFile(file_=str(url.path()), entry=None)

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            qt.QListWidget.dragEnterEvent(self, event)

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            qt.QListWidget.dragMoveEvent(self, event)


class SerieManualControlDialog(qt.QDialog):
    """
    Same as the :class:`SerieManualDefinitionDialog` but with control of the serie.
    This include a `create serie` and a `create serie and clear button`
    """

    sigSerieSend = qt.Signal(Serie)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = SerieManualDefinitionDialog(parent=self)
        self._mainWidget.setWindowFlags(qt.Qt.Widget)
        self.layout().addWidget(self._mainWidget)

        self._buttons = qt.QDialogButtonBox(parent=self)
        self._createButton = qt.QPushButton("create serie", parent=self)
        self._buttons.addButton(self._createButton, qt.QDialogButtonBox.ActionRole)

        # connect signal / slot
        self._createButton.clicked.connect(self._sendSerie)
        self.layout().addWidget(self._buttons)

        # expose API
        self.sigUpdated = self._mainWidget.sigUpdated

    def _sendSerie(self):
        self.sigSerieSend.emit(self._mainWidget.getSerie())

    @property
    def n_tomo_objs(self):
        return self._mainWidget.n_tomo_objs

    def setSerieName(self, name: str):
        self._mainWidget.setSerieName(name=name)

    def getSerieName(self) -> str:
        return self._mainWidget.getSerieName()

    def setSerie(self, serie: Serie) -> None:
        self._mainWidget.setSerie(serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._mainWidget.getSerie(*args, **kwargs)

    def addScanFromNxFile(self, file_: str, entry: Optional[str] = None):
        return self._mainWidget.addScanFromNxFile(file_=file_, entry=entry)

    def removeSelectedScans(self) -> None:
        return self._mainWidget.removeSelectedTomoObjs()

    def getSelectedScans(self) -> tuple:
        return self._mainWidget.getSelectedTomoObjs()

    def setSelectedScans(self, scans: Iterable) -> None:
        self._mainWidget.setSelectedTomoObjs(scans=scans)

    def addScan(self, scan: TomwerScanBase) -> None:
        self._mainWidget.addTomoObj(scan=scan)

    def removeScan(self, scan: TomwerScanBase) -> None:
        self._mainWidget.removeTomoObj(scan=scan)

    def clearSerie(self) -> None:
        self._mainWidget.clearSerie()

    def createSerie(self):
        self.sigSerieSend.emit(self.getSerie())


class SerieManualFromTomoObj(qt.QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        style = qt.QApplication.style()
        self.setLayout(qt.QGridLayout())

        self._tomoObjList = TomoObjList(self)
        self.layout().addWidget(self._tomoObjList, 0, 0, 4, 2)

        # right arrow
        self._rightArrowButton = qt.QPushButton(self)
        rightArrowIcon = style.standardIcon(qt.QStyle.SP_ArrowRight)
        self._rightArrowButton.setIcon(rightArrowIcon)
        self._rightArrowButton.setFixedWidth(30)
        self.layout().addWidget(self._rightArrowButton, 1, 2, 1, 1)

        # left arrow
        self._leftArrowButton = qt.QPushButton(self)
        leftArrowIcon = style.standardIcon(qt.QStyle.SP_ArrowLeft)
        self._leftArrowButton.setIcon(leftArrowIcon)
        self._leftArrowButton.setFixedWidth(30)
        self.layout().addWidget(self._leftArrowButton, 2, 2, 1, 1)

        # new serie
        self._newSerieWidget = NewSerieWidget(self)
        self.layout().addWidget(self._newSerieWidget, 0, 3, 4, 2)

        # tomo obj details
        self._tomoObjInfos = TomoObjOverview(self)
        self._tomoObjInfos.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._tomoObjInfos, 4, 0, 2, 2)

        # connect signals / slot
        self._leftArrowButton.released.connect(self._removeSelectedObjs)
        self._rightArrowButton.released.connect(self._addSelectedObjs)
        self._tomoObjList.selectionModel().selectionChanged.connect(
            self._updateTomoObjInfos
        )

    def selectedTomoObjects(self) -> tuple:
        """
        :return: tuple of tomo object selected on the list
        :rtype: tuple
        """
        items = self._tomoObjList.selectedItems()
        return [item.data(qt.Qt.UserRole) for item in items]

    def _removeSelectedObjs(self, *args, **kwargs):
        for tomo_obj in self.selectedTomoObjects():
            self._newSerieWidget.removeTomoObjToCurrentSerie(tomo_obj)

    def _addSelectedObjs(self, *args, **kwargs):
        for tomo_obj in self.selectedTomoObjects():
            self._newSerieWidget.addTomoObjToCurrentSerie(tomo_obj)

    def _updateTomoObjInfos(self, *args, **kwargs):
        # should
        select_objs = self._tomoObjList.selectedItems()
        if select_objs and len(select_objs) > 0:
            tomo_obj = select_objs[0].data(qt.Qt.UserRole)
            self._tomoObjInfos.setTomoObj(tomo_obj)
        else:
            self._tomoObjInfos.setTomoObj(None)

    # expose API
    def setSerie(self, serie: Serie):
        self._newSerieWidget.setSerie(serie=serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._newSerieWidget.getSerie(*args, **kwargs)

    def addTomoObj(self, tomo_obj):
        self._tomoObjList.add(tomo_obj)

    def addToCurrentSerie(self, tomo_obj):
        self._newSerieWidget.addTomoObjToCurrentSerie(tomo_obj)

    def setSerieName(self, name: str):
        self._newSerieWidget.setSerieName(name=name)


class NewSerieWidget(qt.QWidget):
    sigNameChanged = qt.Signal()
    """Emit when serie name changed"""

    sigUpdated = qt.Signal()
    """
    Emit when the serie has been updated by the tree
    """

    DEFAULT_SERIE_NAME = "my_serie"

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self._nameWidget = qt.QWidget(self)
        self._nameWidget.setLayout(qt.QHBoxLayout())
        self._nameWidget.layout().addWidget(qt.QLabel("serie name", self))
        self._nameQLE = qt.QLineEdit(self.DEFAULT_SERIE_NAME, self)
        self._nameWidget.layout().addWidget(self._nameQLE)
        self.layout().addWidget(self._nameWidget)

        self._serieTree = _SerieDefinitionTree(self, serie_name=self.DEFAULT_SERIE_NAME)
        self.layout().addWidget(self._serieTree)

        # Signal / slot connection
        self._serieTree.itemChanged.connect(self._handleItemUpdate)
        self._serieTree.sigUpdated.connect(self._repeatUpdateSignal)
        self._nameQLE.textChanged.connect(self._nameChangedOnQLE)

    def setSerie(self, serie: Serie) -> None:
        with blockSignals(self._nameQLE):
            self._nameQLE.setText(serie.name)
        self._serieTree.setSerie(serie=serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._serieTree.getSerie(*args, **kwargs)

    def _nameChangedOnQLE(self, name):
        with blockSignals(self._serieTree):
            self._serieTree.setSerieName(name)
        self.sigNameChanged.emit()

    def _handleItemUpdate(self, item, column):
        if item == self._serieTree.rootItem:
            old = self.blockSignals(True)
            self._nameQLE.setText(self._serieTree.rootItem.text(0))
            self.blockSignals(old)
        self.sigUpdated.emit()

    def _repeatUpdateSignal(self):
        self.sigUpdated.emit()

    def addTomoObjToCurrentSerie(self, tomo_obj: TomwerObject):
        assert isinstance(
            tomo_obj, TomwerObject
        ), f"invalid type {type(tomo_obj)}. {TomwerObject} expected"
        self._serieTree.addTomoObj(tomo_obj)

    def removeTomoObjToCurrentSerie(self, tomo_obj: TomwerObject):
        assert isinstance(
            tomo_obj, TomwerObject
        ), f"invalid type {type(tomo_obj)}. {TomwerObject} expected"
        self._serieTree.removeTomoObj(tomo_obj)

    def getSerieName(self) -> str:
        return self._serieTree.getSerieName()

    def setSerieName(self, name: str):
        self._serieTree.setSerieName(name=name)


class SerieManualDefinitionDialog(qt.QDialog):
    """Dialog to define a serie manually"""

    sigUpdated = qt.Signal()

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self._newSerieWidget = NewSerieWidget(self)

        self._buttons = qt.QDialogButtonBox(parent=self)

        self._addScanButton = qt.QPushButton("Add scan to the serie", parent=self)
        self._buttons.addButton(self._addScanButton, qt.QDialogButtonBox.ActionRole)

        self._removeSelectedButton = qt.QPushButton(
            "Remove selected scans", parent=self
        )
        self._buttons.addButton(
            self._removeSelectedButton, qt.QDialogButtonBox.ActionRole
        )

        self._clearButton = qt.QPushButton("Clear", parent=self)
        self._buttons.addButton(self._clearButton, qt.QDialogButtonBox.ActionRole)

        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._newSerieWidget.sigNameChanged.connect(self._nameChanged)
        self._newSerieWidget.sigUpdated.connect(self._repeatUpdateSignal)
        self._addScanButton.clicked.connect(self.addScanFromFileDialog)
        self._removeSelectedButton.clicked.connect(self.removeSelectedTomoObjs)
        self._clearButton.clicked.connect(self.clearSerie)

    @property
    def n_tomo_objs(self):
        serieTree = self._newSerieWidget._serieTree
        return serieTree.n_tomo_objs

    def _nameChanged(self, new_name):
        serieTree = self._newSerieWidget._serieTree
        with blockSignals(self._serieTree):
            serieTree.setSerieName(name=new_name)
        self.sigUpdated.emit()

    def _repeatUpdateSignal(self):
        self.sigUpdated.emit()

    def setSerieName(self, name: str):
        self._newSerieWidget.setSerieName(name=name)

    def getSerieName(self) -> str:
        return self._newSerieWidget.getSerieName()

    def setSerie(self, serie: Serie) -> None:
        self._newSerieWidget.setSerie(serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._newSerieWidget.getSerie(*args, **kwargs)

    def addScanFromFileDialog(self) -> None:
        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        foldersSelected = dialog.files_selected()
        for folder in foldersSelected:
            self.addScanFromNxFile(file_=folder, entry=None)

    def addScanFromNxFile(self, file_: str, entry: Optional[str] = None):
        serieTree = self._newSerieWidget._serieTree
        return serieTree.addScanFromNxFile(file_=file_, entry=entry)

    def removeSelectedTomoObjs(self) -> None:
        serieTree = self._newSerieWidget._serieTree
        return serieTree.removeSelectedTomoObjs()

    def getSelectedTomoObjs(self) -> tuple:
        serieTree = self._newSerieWidget._serieTree
        return serieTree.getSelectedTomoObjs()

    def setSelectedTomoObjs(self, scans: Iterable) -> None:
        serieTree = self._newSerieWidget._serieTree
        serieTree.setSelectedTomoObjs(objs=scans)

    def addTomoObj(self, scan: TomwerScanBase) -> None:
        serieTree = self._newSerieWidget._serieTree
        return serieTree.addTomoObj(tomo_obj=scan)

    def removeTomoObj(self, scan: TomwerScanBase) -> None:
        serieTree = self._newSerieWidget._serieTree
        serieTree.removeTomoObj(tomo_obj=scan)

    def clearSerie(self) -> None:
        serieTree = self._newSerieWidget._serieTree
        serieTree.clearTomoObjs()


class SerieAutomaticDefinitionWidget(qt.QWidget):
    pass


class SerieTree(qt.QWidget):
    """
    Widget used to define a scan serie from a list of scans.
    """

    def __init__(self, parent=None, scans=tuple()) -> None:
        self._series = {}
        # associated serie name (key) to tuple (serie, QTreeWidgetItem)
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._tree = qt.QTreeWidget(self)
        self._tree.setSelectionMode(qt.QAbstractItemView.MultiSelection)
        self._tree.setColumnCount(2)
        self._tree.setHeaderLabels(("serie", "scan ids"))
        self.layout().addWidget(self._tree)

        # set up
        [self.addSerie(scan) for scan in scans]

    def addSerie(self, serie: Serie):
        if not isinstance(serie, Serie):
            raise TypeError(
                f"{serie} is expected to be an instance of {Serie} not {type(serie)}"
            )
        if serie.name in self._series:
            self.removeSerie(self._series[serie.name][0])

        root_item = qt.QTreeWidgetItem(self._tree)
        root_item.setText(0, serie.name)
        self._series[serie.name] = (serie, root_item)
        for obj in serie:
            scan_item = qt.QTreeWidgetItem(root_item)
            if isinstance(obj, TomwerObject):
                text = obj.get_identifier().to_str()
            else:
                text = obj
            scan_item.setText(1, text)
            scan_item.setFlags(qt.Qt.NoItemFlags)

    def removeSerie(self, serie: Serie):
        if not isinstance(serie, Serie):
            raise TypeError(
                f"{serie} is expected to be an instance of {Serie} not {type(serie)}"
            )
        if serie.name in self._series:
            _, serie_item = self._series.pop(serie.name)
            root = self._tree.invisibleRootItem()
            root.removeChild(serie_item)

    @property
    def n_series(self):
        return len(self._series)

    def series(self) -> tuple:
        series = []
        [series.append(serie) for serie, _ in self._series.values()]
        return tuple(series)

    def clearSelection(self):
        self._tree.selectionModel().clearSelection()

    def setSelectedSeries(self, series):
        self.clearSelection()
        for serie in series:
            serie_item = self._getSerieItem(serie)
            if serie_item is not None:
                serie_item.setSelected(True)

    def _getSerieItem(self, serie: Serie) -> Optional[qt.QTreeWidgetItem]:
        if not isinstance(serie, Serie):
            raise TypeError(
                f"serie is expected to be an instance of {Serie} not {type(serie)}"
            )
        return self._series.get(serie.name, (None, None))[1]

    def getSelectedSeries(self) -> tuple():
        """return selected series"""
        selected = []
        for _, (serie, item) in self._series.items():
            if item.isSelected():
                selected.append(serie)
        return tuple(selected)

    def removeSelected(self) -> None:
        for serie in self.getSelectedSeries():
            self.removeSerie(serie)


class SerieHistoryDialog(qt.QDialog):
    sigSerieSend = qt.Signal(Serie)
    """signal emit when a serie has been selected by the user"""

    sigEditSerie = qt.Signal(Serie)
    """Signal emit when user request to edit a serie"""

    sigHistoryUpdated = qt.Signal()
    """Signal emit when the history has been modified"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self._serieList = SerieTree(self)
        self.layout().addWidget(self._serieList)

        self._buttons = qt.QDialogButtonBox(parent=self)

        self._editButton = qt.QPushButton("Edit", parent=self)
        self._buttons.addButton(self._editButton, qt.QDialogButtonBox.ActionRole)

        self._sendButton = qt.QPushButton("Resend", parent=self)
        self._buttons.addButton(self._sendButton, qt.QDialogButtonBox.ActionRole)

        self._removeButton = qt.QPushButton("Remove", parent=self)
        self._buttons.addButton(self._removeButton, qt.QDialogButtonBox.ActionRole)

        self._clearButton = qt.QPushButton("Clear", parent=self)
        self._buttons.addButton(self._clearButton, qt.QDialogButtonBox.ActionRole)

        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._sendButton.clicked.connect(self.sendSelected)
        self._removeButton.clicked.connect(self.removeSelected)
        self._clearButton.clicked.connect(self.clearSelection)
        self._editButton.clicked.connect(self.editSelected)

    def addSerie(self, serie: Serie):
        old = self.blockSignals(True)
        self._serieList.addSerie(serie)
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def removeSerie(self, serie: Serie):
        old = self.blockSignals(True)
        self._serieList.removeSerie(serie)
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def getSelectedSeries(self):
        return self._serieList.getSelectedSeries()

    def setSelectedSeries(self, series):
        old = self.blockSignals(True)
        self._serieList.setSelectedSeries(series)
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def sendSelected(self):
        for serie in self.getSelectedSeries():
            self.sigSerieSend.emit(serie)

    def editSelected(self):
        selected = self.getSelectedSeries()
        if len(selected) == 0:
            return
        if len(selected) > 1:
            _logger.warning(
                "more than one serie selected for edition. Will only edit the first one"
            )
        self.sigEditSerie.emit(selected[0])

    def removeSelected(self):
        old = self.blockSignals(True)
        self._serieList.removeSelected()
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def clearSelection(self):
        self._serieList.clearSelection()

    def series(self) -> tuple:
        return self._serieList.series()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/02/2021"


import unittest
from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
    SADeltaBetaProcess,
    SADeltaBetaParams,
)
from tomwer.core.utils.scanutils import MockHDF5
import os
import tempfile
import shutil


class TestSADeltaBetaProcess(unittest.TestCase):
    """Test the SAAxisProcess class"""

    def setUp(self) -> None:
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockHDF5(
            scan_path=self.tempdir, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        self.scan = mock.scan

    def tearDown(self) -> None:
        shutil.rmtree(self.tempdir)
        super().tearDown()

    def test(self):
        process = SADeltaBetaProcess(inputs={"data": self.scan})

        default_sadelta_beta_params = SADeltaBetaParams()
        default_sadelta_beta_params.output_dir = os.path.join(
            self.tempdir, "output_dir"
        )
        default_sadelta_beta_params.dry_run = True

        with self.assertRaises(Exception):
            process.run()
        process.set_configuration(default_sadelta_beta_params)
        process.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import argparse
from silx.gui import qt
import signal
from typing import Union
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.core.scan.scanfactory import ScanFactory
from silx.io.utils import h5py_read_dataset
from tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta import (
    SADeltaBetaProcess,
)
from tomwer.gui.reconstruction.sadeltabeta.saadeltabeta import (
    SADeltaBetaWindow as _SADeltaBetaWindow,
)
from tomwer.synctools.axis import QAxisRP
from tomwer.gui.reconstruction.saaxis.saaxis import NabuAutoCorDiag
from tomwer.core.process.reconstruction.axis.axis import AxisProcess
from tomwer.core.process.reconstruction.axis.axis import NoAxisUrl
from tomwer.synctools.sadeltabeta import QSADeltaBetaParams
from tomwer.core.process.task import Task
from tomwer.io.utils.h5pyutils import EntryReader
from tomwer.core.process import utils as core_utils
from tomwer.core.process.reconstruction.darkref.darkrefs import (
    requires_reduced_dark_and_flat,
)
from tomwer.core.utils.resource import increase_max_number_file
import logging
import time


logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class SADeltaBetaThread(qt.QThread):
    """
    Thread to call nabu and reconstruct one slice with several cor value
    """

    def init(self, data, configuration, dump_roi):
        self.scan = data
        self._configuration = configuration
        self._dump_roi = dump_roi

    def run(self) -> None:
        process = SADeltaBetaProcess(process_id=None, inputs={"scan": self.scan})
        process.set_configuration(self._configuration)
        process.dump_roi = self._dump_roi
        t0 = time.time()
        process.run()
        print("execution time is {}".format(time.time() - t0))


class SADeltaBetaWindow(_SADeltaBetaWindow):
    def __init__(self, parent=None, dump_roi=False):
        self._scan = None
        super().__init__(parent)
        self._insert_cor_gui()
        # thread for computing cors
        self._processingThread = SADeltaBetaThread()
        self._processingThread.finished.connect(self._threadedProcessEnded)
        # thread to compute the COR
        self._dump_roi = dump_roi

        # hide the validate button
        self._sadbControl._applyBut.hide()
        self.hideAutoFocusButton()

    def _insert_cor_gui(self):
        self._imageWidth = None
        self._automaticCorWidget = None
        self._qaxis_rp = QAxisRP()
        # dialog used to compute automatically the cor

        widget = self._tabWidget._deltaBetaSelectionWidget
        self._corGB = qt.QGroupBox(self)
        self._corGB.setTitle("center of rotation")
        self._corGB.setLayout(qt.QFormLayout())
        # auto cor calculation
        self._autoPB = qt.QPushButton("auto", self)
        self._corGB.layout().addRow(self._autoPB)

        # relative position
        self._relativeCORValueQLE = qt.QDoubleSpinBox(self)
        self._relativeCORValueQLE.setRange(-99999999, 99999999)
        self._corGB.layout().addRow("cor relative position", self._relativeCORValueQLE)
        # absolute position
        self._absoluteCORValueQLE = qt.QDoubleSpinBox(self)
        self._corGB.layout().addRow("cor absolute position", self._absoluteCORValueQLE)
        self._absoluteCORValueQLE.setRange(0, 99999999)
        widget.layout().insertWidget(2, self._corGB)

        # connect signal / slot
        self._relativeCORValueQLE.editingFinished.connect(self._relativeValueEdited)
        self._absoluteCORValueQLE.editingFinished.connect(self._absolueValueEdited)
        self._autoPB.pressed.connect(self._autoCorRequested)

    def _autoCorRequested(self):
        window = self.getAutomaticCorWindow()
        window.activateWindow()
        window.raise_()
        window.show()

    def getAutomaticCorWindow(self):
        if self._automaticCorWidget is None:
            self._automaticCorWidget = NabuAutoCorDiag(self, qarixrp=self._qaxis_rp)
            self._automaticCorWidget.setWindowTitle(
                "compute estimated center of rotation"
            )
            auto_cor_icon = icons.getQIcon("a")
            self._automaticCorWidget.setWindowIcon(auto_cor_icon)
            self._automaticCorWidget.sigRequestAutoCor.connect(
                self._computeEstimatedCor
            )
        return self._automaticCorWidget

    def setScan(self, scan):
        self._imageWidth = scan.dim_1
        # force update of the cor value
        self._relativeCORValueQLE.editingFinished.emit()
        super().setScan(scan=scan)

    def _relativeValueEdited(self):
        old = self._absoluteCORValueQLE.blockSignals(True)
        value = self._relativeCORValueQLE.value() + self._imageWidth / 2.0
        self._absoluteCORValueQLE.setValue(value)
        self._absoluteCORValueQLE.blockSignals(old)

    def _absolueValueEdited(self):
        old = self._relativeCORValueQLE.blockSignals(True)
        value = self._absoluteCORValueQLE.value() - self._imageWidth / 2.0
        self._relativeCORValueQLE.setValue(value)
        self._relativeCORValueQLE.blockSignals(old)

    def setRelativeCorPosition(self, value):
        self._relativeCORValueQLE.setValue(value)
        self._relativeCORValueQLE.editingFinished.emit()

    def getAbsoluteCorPosition(self):
        return self._absoluteCORValueQLE.value()

    def _launchReconstructions(self):
        if self._processingThread.isRunning():
            _logger.error(
                "a calculation is already launch. You must wait for "
                "it to end prior to launch a new one"
            )
        else:
            qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
            self._processingThread.init(
                configuration=self.getConfiguration(),
                data=self.getScan(),
                dump_roi=self._dump_roi,
            )
            self._processingThread.start()

    def getConfiguration(self) -> dict:
        config = super().getConfiguration()
        # insert rotation_axis_position
        if "reconstruction" not in config["nabu_params"]:
            config["nabu_params"]["reconstruction"] = {}
        config["nabu_params"]["reconstruction"][
            "rotation_axis_position"
        ] = self.getAbsoluteCorPosition()
        return config

    def _threadedProcessEnded(self):
        sa_delta_beta_params = self._processingThread.scan.sa_delta_beta_params
        if sa_delta_beta_params is None:
            scores = None
        else:
            scores = sa_delta_beta_params.scores
        scan = self.getScan()
        assert scan is not None, "scan should have been set"
        self.setDBScores(
            scores, img_width=scan.dim_1, score_method=self.getScoreMethod()
        )
        if scan.sa_delta_beta_params.autofocus is not None:
            self.setCurrentDeltaBetaValue(scan.sa_delta_beta_params.autofocus)
        qt.QApplication.restoreOverrideCursor()
        self.showResults()

    def _stopProcessingThread(self):
        if self._processingThread:
            self._processingThread.terminate()
            self._processingThread.wait(500)
            self._processingThread = None

    def stop(self):
        self._stopProcessingThread()
        super().stop()

    def _computeEstimatedCor(self) -> Union[float, None]:
        scan = self.getScan()
        if scan is None:
            return
        _cor_estimation_process = AxisProcess(
            inputs={"data": scan, "axis_params": self._qaxis_rp}
        )

        _logger.info("{} - start cor estimation for".format(scan))
        qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
        try:
            _cor_estimation_process.compute(scan=scan, wait=True)
        except NoAxisUrl:
            qt.QApplication.restoreOverrideCursor()
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = (
                "Unable to find url to compute the axis, please select them "
                "from the `axis input` tab"
            )
            msg.setText(text)
            msg.exec_()
            return None
        else:
            self.setRelativeCorPosition(
                value=scan.axis_params.relative_cor_value,
            )
            qt.QApplication.restoreOverrideCursor()
            self.getAutomaticCorWindow().hide()
            return scan.axis_params.relative_cor_value


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus "
        "provide the master file",
        default=None,
    )
    parser.add_argument(
        "--entry",
        help="For Nexus files: entry in the master file",
        default=None,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    parser.add_argument(
        "--read-existing",
        dest="read_existing",
        action="store_true",
        default=False,
        help="Load latest sa-delta-beta processing from *_tomwer_processes.h5 "
        "if exists",
    )
    parser.add_argument(
        "--dump-roi",
        dest="dump_roi",
        action="store_true",
        default=False,
        help="Save roi where the score is computed on the .hdf5",
    )
    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()
    scan = ScanFactory.create_scan_object(
        scan_path=options.scan_path, entry=options.entry
    )
    requires_reduced_dark_and_flat(scan=scan, logger_=_logger)

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    window = SADeltaBetaWindow(dump_roi=options.dump_roi)
    window.setWindowTitle("sa-delta-beta")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    if not hasattr(scan, "sa_delta_beta_params") or scan.sa_delta_beta_params is None:
        scan.sa_delta_beta_params = QSADeltaBetaParams()
    window.setScan(scan)
    if options.read_existing is True:
        scores, selected = _load_latest_scores(scan)
        if scores is not None:
            window.setDBScores(scores, score_method="standard deviation")
            if selected not in (None, "-"):
                window.setCurrentDeltaBetaValue(selected)

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.aboutToQuit.connect(window.stop)
    app.exec_()


def getinputinfo():
    return "tomwer saaxis [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def _load_latest_scores(scan) -> tuple:
    """

    :param scan:
    :return: loaded scores and selected score as (scores, selected).
             "scores" can be None or a dict. "selected" can be None or a float
    """
    scores = None
    selected = None
    if scan.process_file is None:
        _logger.warning(
            "Unable to find process file. Unable to read " "existing processing"
        )
        return scores, selected

    with EntryReader(scan.process_file_url) as h5f:
        latest_sa_db_node = Task.get_most_recent_process(h5f, SADeltaBetaProcess)
        if latest_sa_db_node and "results" in latest_sa_db_node:
            scores = core_utils.get_scores(latest_sa_db_node)
            if "delta_beta" in latest_sa_db_node["results"]:
                selected = h5py_read_dataset(latest_sa_db_node["results"]["delta_beta"])
        else:
            _logger.warning("no results found for {}".format(scan))
    return scores, selected


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from ..utils import WidgetLongProcessing
from silx.gui import qt
from orangewidget import gui
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.darkref.darkrefcopywidget import DarkRefAndCopyWidget
from tomwer.synctools.darkref import QDKRFRP
from tomwer.synctools.stacks.reconstruction.dkrefcopy import DarkRefCopyProcessStack
from processview.core.manager import DatasetState
from processview.core.manager import ProcessManager
import tomoscan.esrf.scan.utils
from tomwer.utils import docstring
import tomwer.core.process.reconstruction.darkref.darkrefscopy
import copy

_logger = logging.getLogger(__name__)


class DarkRefAndCopyOW(SuperviseOW, WidgetLongProcessing):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "reduced darks and flats"
    id = "orange.widgets.tomwer.darkrefs"
    description = (
        "This widget will generate reduced darks and flats for a received scan "
    )
    icon = "icons/darkref.svg"
    priority = 15
    keywords = ["tomography", "dark", "darks", "ref", "refs", "flat", "flats"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _ewoks_default_inputs = Setting({"data": None, "dark_ref_params": None})

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    ewokstaskclass = (
        tomwer.core.process.reconstruction.darkref.darkrefscopy.DarkRefsCopy
    )

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, doc="one scan to be process")
        reduced_darks = Input(
            name="reduced dark(s)", type=dict, doc="dict containing reduced dark(s)"
        )
        reduced_flats = Input(
            name="reduced flat(s)",
            type=dict,
            doc="dict of containing reduced flat(s)",
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")
        reduced_darks = Output(
            name="reduced dark(s)",
            type=dict,
            doc="DataUrl of the reduced dark. Key is the relative position, value is the reduced dark",
        )
        reduced_flats = Output(
            name="reduced flat(s)",
            type=dict,
            doc="DataUrl of the reduced flat. Key is the relative position, value is the reduced flat",
        )

    def __init__(self, parent=None, reconsparams=None):
        """

        :param QKRFRP reconsparams: reconstruction parameters
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        recons_params = reconsparams or QDKRFRP()
        self._processing_stack = DarkRefCopyProcessStack(process_id=self.process_id)

        dark_ref_params = self._ewoks_default_inputs.get("dark_ref_params", None)
        if dark_ref_params not in ({}, None):
            try:
                recons_params.dkrf.load_from_dict(dark_ref_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        self.widget = DarkRefAndCopyWidget(
            save_dir=self._processing_stack._save_dir,
            parent=self,
            reconsparams=recons_params,
            process_id=self.process_id,
        )
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self.widget)
        self.setForceSync = self.widget.setForceSync

        # expose API
        self.setModeAuto = self.widget.set_mode_auto
        self.setRefsFromScan = self.widget.setRefsFromScan
        self.setCopyActive = self.widget.setCopyActive

        # connect signal / slot
        self.widget.sigProcessingStart.connect(self._startProcessing)
        self.widget.sigProcessingEnd.connect(self._endProcessing)
        self.widget.sigScanReady.connect(self.signalReady)
        self.widget.recons_params.sigChanged.connect(self._updateSettingsVals)
        self.widget.sigModeAutoChanged.connect(self._updateSettingsVals)
        self.widget.sigCopyActivationChanged.connect(self._updateSettingsVals)
        self._processing_stack.sigComputationStarted.connect(self._startProcessing)
        self._processing_stack.sigComputationEnded.connect(self._endProcessing)
        self._processing_stack.sigRefSetted.connect(self.widget.setRefSetBy)

        # load some other copy parameters
        if dark_ref_params not in ({}, None):
            try:
                if "activate" in dark_ref_params:
                    self.widget.setCopyActive(dark_ref_params.pop("activate"))
                if "auto" in dark_ref_params:
                    auto_mode = dark_ref_params.pop("auto")
                    # insure backward compatibility. Has beem saved as a tuple (this was a typo)
                    if isinstance(auto_mode, tuple) and isinstance(auto_mode[0], bool):
                        auto_mode = auto_mode[0]
                    self.widget.setModeAuto(auto_mode)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        self.setCaption(self.windowTitle())

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process(self, scanID):
        if scanID is None:
            return
        assert isinstance(scanID, TomwerScanBase)
        ProcessManager().notify_dataset_state(
            dataset=scanID, process=self, state=DatasetState.PENDING
        )
        configuration = self.widget.recons_params.to_dict()
        configuration["process_only_copy"] = False  # we never only do copy from the gui
        configuration["process_only_dkrf"] = not self.widget.isCopyActive()

        self._processing_stack.add(copy.copy(scanID), configuration=configuration)

    @Inputs.reduced_darks
    def received_reduced_darks(self, reduced_darks: dict):
        # we consider if the darks are provided then the user want to use those and don't copy them manually
        self.setModeAuto(False)
        self.setReducedDarks(reduced_darks)

    @Inputs.reduced_flats
    def received_reduced_flats(self, reduced_flats: dict):
        # we consider if the darks are provided then the user want to use those and don't copy them manually
        self.setModeAuto(False)
        self.setReducedFlats(reduced_flats)

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)

    def signalReady(self, scanID):
        assert isinstance(scanID, TomwerScanBase)
        self.Outputs.data.send(scanID)
        self.sigScanReady.emit(scanID)

    def _updateSettingsVals(self):
        self._ewoks_default_inputs = {
            "data": None,
            "dark_ref_params": self.widget.recons_params.to_dict(),
        }
        self._ewoks_default_inputs["dark_ref_params"][
            "auto"
        ] = self.widget.isOnModeAuto()
        self._ewoks_default_inputs["dark_ref_params"][
            "activate"
        ] = self.widget.isCopyActive()

    @property
    def recons_params(self):
        return self.widget.recons_params

    def close(self):
        self.widget.close()
        super(DarkRefAndCopyOW, self).close()

    def _endProcessing(self, scan):
        WidgetLongProcessing._endProcessing(self, scan)
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)
        if scan.reduced_darks not in (None, {}):
            # we want to send those in relative position to have something generic. This is a convention for now
            self.Outputs.reduced_darks.send(
                tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                    reduced_frames=scan.reduced_darks, scan=scan
                )
            )
        if scan.reduced_flats not in (None, {}):
            # we want to send those in relative position to have something generic. This is a convention for now
            self.Outputs.reduced_flats.send(
                tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                    reduced_frames=scan.reduced_flats, scan=scan
                )
            )
        _logger.info("{} ended".format(str(scan)))

    def setReducedDarks(self, darks: dict):
        self.widget._refCopyWidget.save_darks_to_be_copied(darks)

    def setReducedFlats(self, flats: dict):
        self.widget._refCopyWidget.save_flats_to_be_copied(flats)



#!/usr/bin/python
# coding: utf-8
#
#    Project: Azimuthal integration
#             https://github.com/pyFAI/pyFAI
#
#    Copyright (C) 2015 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


__doc__ = """test module for pyFAI."""
__authors__ = ["Jérôme Kieffer", "Valentin Valls", "Henri Payno"]
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "07/02/2017"

import os
import shutil
import fileinput
from tomwer.core.utils import DownloadDataset, url_base
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class UtilsTest(object):
    """
    Static class providing useful stuff for preparing tests.
    """

    timeout = 100  # timeout in seconds for downloading datasets.tar.bz2

    def __init__(self):
        self.installed = False

    @classmethod
    def dataDownloaded(cls, archive_folder, archive_file):
        return cls.dataIsHere(archive_folder=archive_folder) or cls.dataIsDownloaded(
            archive_file=archive_file
        )

    @classmethod
    def dataIsHere(cls, archive_folder):
        return os.path.isdir(archive_folder)

    @classmethod
    def dataIsDownloaded(cls, archive_file):
        return os.path.isfile(archive_file)

    @classmethod
    def getEDFDataset(cls, folderID):
        path = os.path.abspath(
            os.path.join(cls.getDatasets(name="edf_datasets"), folderID)
        )
        if os.path.isdir(path):
            return path
        else:
            raise RuntimeError("Coul'd find folder containing scan %s" % folderID)

    @classmethod
    def getH5Dataset(cls, folderID):
        path = os.path.abspath(
            os.path.join(cls.getDatasets(name="h5_datasets"), folderID)
        )
        if os.path.exists(path):
            return path
        else:
            raise RuntimeError("Coul'd find folder containing scan %s" % folderID)

    @classmethod
    def getBlissDataset(cls, folderID):
        path = os.path.abspath(
            os.path.join(cls.getDatasets(name="bliss_acq"), folderID)
        )
        if os.path.exists(path):
            return path
        else:
            raise RuntimeError("Coul'd find folder containing scan %s" % folderID)

    @classmethod
    def getRecDatasets(cls):
        path = os.path.abspath(
            cls.getDatasets(
                name="rec_HA-700_45.4_136keV_BM5_burrow_GSA-1-44_021_avg-2_"
            )
        )
        if os.path.exists(path):
            return path
        else:
            raise RuntimeError("Coul'd find folder containing rec")

    @classmethod
    def getOrangeTestFile(cls, folderID):
        path = os.path.abspath(os.path.join(cls.getOrangeTestFiles(), folderID))
        if os.path.isfile(path):
            return path
        else:
            raise RuntimeError("Coul'd find folder containing scan %s" % folderID)

    @classmethod
    def getOrangeTestFiles(cls):
        return cls.getDatasets(name="orangetestfiles")

    @classmethod
    def getOrangeExecTestFile(cls, folderID):
        path = os.path.abspath(os.path.join(cls.getOrangeExecTestFiles(), folderID))
        if os.path.isfile(path):
            return path
        else:
            raise RuntimeError("Coul'd find folder containing scan %s" % folderID)

    @classmethod
    def getOrangeExecTestFiles(cls):
        return cls.getDatasets(name="ows_test_exec")

    @classmethod
    def getDatasets(cls, name="datasets"):
        """
        Downloads the requested image from Forge.EPN-campus.eu

        @param: name of the image.
        For the RedMine forge, the filename contains a directory name that is removed
        @return: full path of the locally saved file
        """
        archive_file = name + ".tar.bz2"
        archive_folder = "".join((os.path.dirname(__file__), "/" + name + "/"))
        archive_file = os.path.join(archive_folder, archive_file)

        # download if needed
        if not cls.dataDownloaded(
            archive_folder=archive_folder, archive_file=archive_file
        ):
            DownloadDataset(
                dataset=os.path.basename(archive_file),
                output_folder=archive_folder,
                timeout=cls.timeout,
            )

            if not os.path.isfile(archive_file):
                raise RuntimeError(
                    "Could not automatically download test images %s!\n"
                    "If you are behind a firewall, please set both environment "
                    "variable http_proxy and https_proxy. "
                    "This even works under windows ! \n Otherwise please try "
                    "to download the images manually from \n%s/%s"
                    % (archive_file, url_base, archive_file)
                )

        # decompress if needed
        if os.path.isfile(archive_file):
            logger.info("decompressing %s." % archive_file)
            outdir = "".join((os.path.dirname(__file__)))
            shutil.unpack_archive(archive_file, extract_dir=outdir, format="bztar")
            os.remove(archive_file)
        else:
            logger.info("not trying to decompress it")
        return archive_folder

    @classmethod
    def hasInternalTest(cls, dataset):
        """
        The id of the internal test is to have some large scan accessible
        which are stored locally. This should be used only for unit test that
        can be skipped
        """
        if "TOMWER_ADDITIONAL_TESTS_DIR" not in os.environ:
            return False
        else:
            dir = os.path.join(os.environ["TOMWER_ADDITIONAL_TESTS_DIR"], dataset)
        return os.path.isdir(dir)

    @classmethod
    def getInternalTestDir(cls, dataset):
        if cls.hasInternalTest(dataset) is False:
            return None
        else:
            return os.path.join(os.environ["TOMWER_ADDITIONAL_TESTS_DIR"], dataset)


def rebaseParFile(filePath, scanOldPath, scanNewPath):
    """
    Change some variables on the oar file to rebase it like if it was acquire
    in scanNewPath and allow pyhst reconstruction.
    """
    assert os.path.isfile(filePath)

    with fileinput.FileInput(filePath, inplace=True, backup=".bak") as file:
        for line in file:
            print(line.replace(scanOldPath, scanNewPath))


def rebaseAcquisition(src, dst):
    """Copy the acquisition from src to dst by changing the ID of the acquisition
    on all files and sub-folders"""

    def copyDir(_dirSrc, _dirDst, srcID, dstID):
        assert os.path.exists(_dirDst) is False
        os.makedirs(_dirDst)
        for _file in os.listdir(_dirSrc):
            _filePath = os.path.join(_dirSrc, _file)
            if os.path.isdir(_filePath):
                copyDir(
                    _dirSrc=_filePath,
                    _dirDst=os.path.join(_dirDst, _file.replace(srcID, dstID, 1)),
                    srcID=srcID,
                    dstID=dstID,
                )
            else:
                shutil.copyfile(
                    src=_filePath,
                    dst=os.path.join(_dirDst, _file.replace(srcID, dstID, 1)),
                )

    # cp folder then copy files
    assert os.path.isdir(src)
    srcID = os.path.basename(src)
    dstID = os.path.basename(dst)
    copyDir(_dirSrc=src, _dirDst=dst, srcID=srcID, dstID=dstID)


def skip_gui_test():
    return os.environ.get("_TOMWER_NO_GUI_UNIT_TESTS", "False") == "True"


def skip_orange_workflows_test():
    return os.environ.get("_TOMWER_NO_ORANGE_WORKFLOWS_UNIT_TESTS", "False") == "True"



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "23/03/2021"


from silx.gui import qt
from tomwer.gui import icons as tomwer_icons


class NXTomomillParamsAction(qt.QAction):
    """
    Action to display a window with nxtomomill configuration
    """

    def __init__(self, parent):
        icon = tomwer_icons.getQIcon("parameters")

        qt.QAction.__init__(self, icon, "filter configuration", parent)
        self.setToolTip("Open dialog to configure nxtomomill parameters")
        self.setCheckable(False)


class CFGFileActiveLabel(qt.QLabel):
    """Label used to display if the .cfg file is active or not"""

    def __init__(self, parent):
        super().__init__(parent)
        icon = tomwer_icons.getQIcon("cfg_file_inactive")
        self.setToolTip("no valid cfg file provided")
        self.setPixmap(icon.pixmap(self.width(), self.height()))

    def setActive(self, active=True):
        if active is True:
            icon = tomwer_icons.getQIcon("cfg_file_active")
            tooltip = "will use the provided .cfg file"
        else:
            icon = tomwer_icons.getQIcon("cfg_file_inactive")
            tooltip = "will use default configuration"

        self.setPixmap(icon.pixmap(self.width(), self.height()))
        self.setToolTip(tooltip)

    def setInactive(self):
        self.setActive(active=False)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/08/2020"


from silx.gui import qt
from tomwer.gui import icons as tomwer_icons


class MinimalisticConfigurationAction(qt.QAction):
    """
    Action to display the 'minimalistic' options only
    """

    def __init__(self, parent):
        self.__icon = tomwer_icons.getQIcon("minimalistic_user")
        qt.QAction.__init__(self, self.__icon, "minimalistic options", parent)
        self.setIconVisibleInMenu(True)
        self.setCheckable(True)
        self.setToolTip(self.tooltip())

    def tooltip(self):
        return "configuration: minimalistic level limit the number of options"

    def icon(self):
        return self.__icon

    def text(self):
        return "Minimalistic configuration"


class BasicConfigurationAction(qt.QAction):
    """
    Action to display the 'basic' options only
    """

    def __init__(self, parent):
        self.__icon = tomwer_icons.getQIcon("basic_user")
        qt.QAction.__init__(self, self.__icon, "basic options", parent)
        self.setIconVisibleInMenu(True)
        self.setCheckable(True)
        self.setToolTip(self.tooltip())

    def tooltip(self):
        return "configuration: basic level limit the number of options"

    def icon(self):
        return self.__icon

    def text(self):
        return "Basic configuration"


class ExpertConfigurationAction(qt.QAction):
    """
    Action to display the 'advanced' / expert options
    """

    def __init__(self, parent):
        self.__icon = tomwer_icons.getQIcon("advanced_user")
        qt.QAction.__init__(self, self.__icon, "advanced options", parent)
        self.setIconVisibleInMenu(True)
        self.setCheckable(True)
        self.setToolTip(self.tooltip())

    def tooltip(self):
        return (
            "configuration: advanced level give user all the possible \n"
            "option to tune the reconstructions"
        )

    def icon(self):
        return self.__icon

    def text(self):
        return "Advanced configuration"


class FilterAction(qt.QAction):
    """
    Action to activate the filtering from the nabu stage
    """

    def __init__(self, parent):
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_FileDialogContentsView)

        qt.QAction.__init__(self, icon, "filter configuration", parent)
        self.setToolTip(
            "If activated will only display the configuration"
            "for the active nabu phase"
        )
        self.setCheckable(True)
        self.setShortcut(qt.QKeySequence(qt.Qt.Key_F))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui to select a slice in a volume
"""


__authors__ = [
    "H. Payno",
]

__license__ = "MIT"
__date__ = "04/02/2021"


from silx.gui import qt
from silx.gui.plot import PlotWidget
import functools
import numpy


# class SliceSelectorWithMiddle(qt.QWidget):


class SliceSelector(qt.QWidget):
    """
    Allow definition of n slices with a volume slice selector and spin boxes
    :param bool insert: if True add spin boxes at the beginning of the layout
                        otherwise append them
    """

    sigSlicesChanged = qt.Signal()
    """signal emitted when slices value change"""

    def __init__(self, parent=None, insert=True, invert_y_axis=False):
        self._updatePlaneFrmSBLock = False
        self._insert = insert
        self._slicesQSB = []
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QBoxLayout(qt.QBoxLayout.BottomToTop))
        self._volumeView = _SliceSelectorFrmVolume(self, invert_y_axis=invert_y_axis)
        self.layout().addWidget(self._volumeView)

    def setSlicesRange(self, min_index, max_index):
        self._volumeView.setSlicesRange(min_index, max_index)
        for widget in self._slicesQSB:
            widget.sliceSB.setRange(min_index, max_index)

    def clearSlices(self):
        self._volumeView.clearSlices()
        for sliceQSB in self._slicesQSB:
            self.layout().removeWidget(sliceQSB)
        self._slicesQSB.clear()

    def addSlice(self, value, name, color):
        self._volumeView.addSlice(value=value, name=name, color=color)
        # create the spin box
        widget = qt.QWidget(self)
        widget.setLayout(qt.QHBoxLayout())
        widget.layout().setContentsMargins(0, 0, 0, 0)
        widget.layout().addWidget(qt.QLabel(name, self))
        sb = qt.QSpinBox(self)
        stylesheet = "background-color: {}".format(color)
        sb.setStyleSheet(stylesheet)

        min_i, max_i = self.getSlicesRange()
        sb.setRange(min_i, max_i)
        sb.setValue(value)
        widget.sliceSB = sb
        widget.layout().addWidget(sb)

        if self._insert:
            self.layout().insertWidget(0, widget)
        else:
            self.layout().addWidget(widget)
        self._slicesQSB.append(widget)

        # connect signal / slot
        sb.valueChanged.connect(functools.partial(self._updatePlaneFrmSB, name, sb))
        sb.valueChanged.connect(self._slicesChanged)
        marker = self._volumeView._getMarker(self._volumeView.getMarkerName(name))
        marker.sigDragFinished.connect(
            functools.partial(self._updateSBFrmMarker, sb, marker)
        )
        marker.sigDragFinished.connect(self._slicesChanged)
        # TODO: we might want to change the plane position when moving the
        # marker. This would require to add a loop when drag start and stop
        # it when drag is finished

    def _slicesChanged(self):
        if not self._updatePlaneFrmSBLock:
            self.sigSlicesChanged.emit()

    def _updatePlaneFrmSB(self, name: str, sb: qt.QSpinBox):
        """callback when a spin box value change from user input"""
        self._updatePlaneFrmSBLock = True
        old = self._volumeView.blockSignals(True)
        marker = self._volumeView._getMarker(self._volumeView.getMarkerName(name))
        marker.setPosition(x=0, y=self._volumeView.valueToPlotSpace(value=sb.value()))
        marker.sigDragFinished.emit()
        self._volumeView.blockSignals(old)
        self._updatePlaneFrmSBLock = False

    def _updateSBFrmMarker(self, sb: qt.QSpinBox, marker):
        if not self._updatePlaneFrmSBLock:
            position = self._volumeView.plotSpaceToValue(marker.getPosition()[1])
            sb.setValue(position)

    # expose API
    def getSlicesValue(self) -> dict:
        return self._volumeView.getSlicesValue()

    def setSliceValue(self, name: str, value: float):
        self._volumeView.setSliceValue(name=name, value=value)

    def getSlicesRange(self) -> tuple:
        return self._volumeView.getSlicesRange()


class _SliceSelectorFrmVolume(PlotWidget):
    """
    Allow definition of n slices on a volume
    """

    sigSlicesChanged = qt.Signal()
    """signal emitted when slices value change"""

    ORIENTATION = -numpy.pi / 2.0
    LOW_RADIUS = 0.4
    HIGH_RADIUS = 1.0
    HEIGHT = 3.0
    PLOT_MARGINS = 1.0

    def __init__(self, parent=None, invert_y_axis=False):
        self._yAxisInverted = invert_y_axis
        PlotWidget.__init__(self, parent)
        self.setAxesDisplayed(False)
        self._plotVolume()
        self.getXAxis().setLimits(-(self.PLOT_MARGINS + 0.5), self.PLOT_MARGINS + 0.5)
        self.getYAxis().setLimits(-self.PLOT_MARGINS, self.HEIGHT + self.PLOT_MARGINS)
        self._slices = []
        self._slicesRange = (0, 1)

        # Retrieve PlotWidget's plot area widget
        plotArea = self.getWidgetHandle()

        # Set plot area custom context menu
        plotArea.setContextMenuPolicy(qt.Qt.CustomContextMenu)
        plotArea.customContextMenuRequested.connect(self._contextMenu)
        self.setInteractiveMode("select", zoomOnWheel=False)
        self.setPanWithArrowKeys(False)

    def _contextMenu(self, pos):
        """Handle plot area customContextMenuRequested signal.

        :param QPoint pos: Mouse position relative to plot area
        """
        # avoir rest zoom action
        pass

    def setSlicesRange(self, min_index, max_index):
        if not isinstance(min_index, int):
            raise TypeError("Invalid type")
        if not isinstance(max_index, int):
            raise TypeError("Invalid type")
        slice_values = self.getSlicesValue()
        self._slicesRange = (
            float(min(min_index, max_index)),
            float(max(min_index, max_index)),
        )

        for slice_name, slice_value in slice_values.items():
            try:
                self.setSliceValue(name=slice_name, value=slice_value)
            except Exception:
                pass

    def getSlicesRange(self) -> tuple:
        return self._slicesRange

    def _plotVolume(self):
        nbpoints = 60
        angles = numpy.arange(nbpoints) * 2.0 * numpy.pi / nbpoints

        # draw ellipsis
        for y_offset, legend in zip(
            (0, self.HEIGHT), ("lower_ellipsis", "higher_ellipsis")
        ):
            X = self.LOW_RADIUS * numpy.cos(angles) * numpy.cos(
                self.ORIENTATION
            ) - self.HIGH_RADIUS * numpy.sin(angles) * numpy.sin(self.ORIENTATION)
            Y = self.LOW_RADIUS * numpy.cos(angles) * numpy.sin(
                self.ORIENTATION
            ) + self.HIGH_RADIUS * numpy.sin(angles) * numpy.cos(self.ORIENTATION)
            X = list(X)
            X.append(X[0])
            X = numpy.array(X)
            Y = list(Y)
            Y.append(Y[0])
            Y = numpy.array(Y)
            self.addCurve(X, Y + y_offset, legend=legend, color="gray")
        # draw sides
        self.addCurve((-1, -1), (0, self.HEIGHT), legend="left side", color="gray")
        self.addCurve((1, 1), (0, self.HEIGHT), legend="right side", color="gray")

    def addSlice(self, value, name, color):
        """

        :param value:
        :param name:
        """
        value = min(max(value, self._slicesRange[0]), self._slicesRange[1])
        self._addSliceAnchor(value, name, color)
        self._slices.append(name)

    def clearSlices(self):
        self._slices.clear()

    def valueToPlotSpace(self, value):
        value = numpy.clip(value, self._slicesRange[0], self._slicesRange[1])
        if self._yAxisInverted:
            value = self._slicesRange[1] - value
        value = (value - self._slicesRange[0]) / (
            self._slicesRange[1] - self._slicesRange[0]
        )
        return value * self.HEIGHT

    def plotSpaceToValue(self, y):
        value = y / self.HEIGHT
        res = int(
            value * (self._slicesRange[1] - self._slicesRange[0]) + self._slicesRange[0]
        )
        if self._yAxisInverted:
            res = self._slicesRange[1] - res
        return int(res)

    def _updatePlane(self, y, curve_name, color):
        anchor_shift = 0.2
        nbpoints = 4
        angles = numpy.arange(nbpoints) * 2.0 * numpy.pi / nbpoints
        X = self.LOW_RADIUS * numpy.cos(angles) * numpy.cos(
            self.ORIENTATION
        ) - self.HIGH_RADIUS * numpy.sin(angles) * numpy.sin(self.ORIENTATION)
        Y = self.LOW_RADIUS * numpy.cos(angles) * numpy.sin(
            self.ORIENTATION
        ) + self.HIGH_RADIUS * numpy.sin(angles) * numpy.cos(self.ORIENTATION)
        y_min = y + min(Y)
        y_max = y + max(Y)

        x_min = min(X)
        x_max = max(X)
        xs = (
            x_min + anchor_shift,
            x_max + anchor_shift,
            x_max - anchor_shift,
            x_min - anchor_shift,
            x_min + anchor_shift,
        )
        self.addCurve(
            x=xs,
            y=(y_max, y_max, y_min, y_min, y_max),
            legend=curve_name,
            color=color,
            resetzoom=False,
        )
        self.sigSlicesChanged.emit()

    def getMarkerName(self, name: str):
        return "marker_{}".format(name)

    def _addSliceAnchor(self, value, curve_name, color):
        y_value = self.valueToPlotSpace(value=value)
        old = self.blockSignals(True)
        self._updatePlane(y=y_value, curve_name=curve_name, color=color)

        marker_name = self.getMarkerName(curve_name)
        self.addYMarker(y=y_value, color=color, legend=marker_name, draggable=True)
        marker = self._getMarker(marker_name)
        if marker:
            marker.setLineStyle("--")
            marker.sigDragFinished.connect(
                functools.partial(
                    self._updateSlicePlane, marker_name, curve_name, color
                )
            )
        self.blockSignals(old)

    def _updateSlicePlane(self, marker_name, curve_name, color):
        marker = self._getMarker(marker_name)
        if marker:
            if marker.getPosition()[1] < 0:
                marker.setPosition(0, 0)
            elif marker.getPosition()[1] > self.HEIGHT:
                marker.setPosition(0, self.HEIGHT)
            self._updatePlane(
                y=marker.getPosition()[1], curve_name=curve_name, color=color
            )

    def getSlicesValue(self) -> dict:
        """

        :return: dict with slice name as key and value as value
        """
        res = {}
        for slice_ in self._slices:
            marker = self._getMarker(self.getMarkerName(slice_))
            if marker:
                res[slice_] = int(self.plotSpaceToValue(marker.getPosition()[1]))
        return res

    def setSliceValue(self, name: str, value: float):
        if name not in self._slices:
            return
        marker = self._getMarker(self.getMarkerName(name))
        marker.setPosition(0, self.valueToPlotSpace(float(value)))
        marker.sigDragFinished.emit()



from tomwer.core.process.reconstruction.paramsbase import _ReconsParam, _get_db_fromstr
from tomwer.core.process.reconstruction.darkref.params import DKRFRP


def test_paramsbase():
    """
    Test _ReconsParam class
    """
    params = _ReconsParam()
    assert isinstance(params.all_params, list)


def test_get_db_fromstr():
    """
    test '_get_db_fromstr' function
    """
    assert _get_db_fromstr(vals="12") == 12.0
    assert _get_db_fromstr(vals="12.0") == 12.0
    assert _get_db_fromstr(vals="(12.0, )") == 12.0
    assert _get_db_fromstr(vals="(12.0, 13.5)") == (12.0, 13.5)
    assert _get_db_fromstr(vals="[12.0, 13.5]") == (12.0, 13.5)


def test_basic_DKRFRP():
    """
    dummy test for DKRFRP
    """
    recons_params = DKRFRP()
    ddict = recons_params.to_dict()
    DKRFRP.from_dict(ddict)
    assert isinstance(
        recons_params.to_unique_recons_set(),
        tuple,
    )



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/11/2021"


import asyncio
import os
import shutil
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
import tempfile
from tomwer.core.utils.scanutils import MockHDF5
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.gui.cluster.supervisor import FutureTomwerScanObserverWidget


class TestSupervisor(TestCaseQt):
    """Test the supervisor GUI"""

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()

        # set up scans
        self._scans = []
        self._future_tomo_objs = []
        for i in range(5):
            # create scan
            scan = MockHDF5(
                scan_path=os.path.join(self.tempdir, f"scan_test{i}"),
                n_proj=10,
                n_ini_proj=10,
                create_ini_dark=False,
                create_ini_ref=False,
                dim=10,
            ).scan
            self._scans.append(scan)

            # create future
            future = asyncio.Future()
            if i == 1:
                future.set_result(None)
            self._future_tomo_objs.append(
                FutureTomwerObject(
                    tomo_obj=scan,
                    futures=(future,),
                )
            )
        # set up gui
        self._supervisor = FutureTomwerScanObserverWidget()
        self._supervisor.setConvertWhenFinished(False)

    def tearDown(self):
        self._supervisor.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._supervisor.close()
        self._supervisor = None
        shutil.rmtree(self.tempdir)
        self._scans.clear()
        self._future_tomo_objs.clear()
        super().tearDown()

    def test(self):
        for future_tomo_obj in self._future_tomo_objs:
            self._supervisor.addFutureTomoObj(future_tomo_obj=future_tomo_obj)
        self._supervisor.show()



from ewokscore.task import Task as EwoksTask
from tomwer.core.utils.scanutils import data_identifier_to_scan


class _VolumeSelectorPlaceHolder(
    EwoksTask, input_names=["volume"], output_names=["volume"]
):
    def run(self):
        self.outputs.volume = data_identifier_to_scan(self.inputs.volume)



"""This is almost the original orange3 PythonScript widget except that
the icon change (to difference them) and here the input is not a Table
but a TomwerScanBase (use silx.gui.qt instead of AnyQt too)"""

import sys
import os
import code
import keyword
import itertools
import unicodedata
from unittest.mock import patch
from silx.gui import qt
from orangewidget import widget, gui
from orangewidget.utils import itemmodels
from orangewidget.settings import Setting, SettingsHandler
from orangewidget.widget import OWBaseWidget, Input, Output
import tomwer.core.process.script.python
from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomoscan.serie import Serie

__all__ = ["OWPythonScript"]


def text_format(foreground=qt.Qt.black, weight=qt.QFont.Normal):
    fmt = qt.QTextCharFormat()
    fmt.setForeground(qt.QBrush(foreground))
    fmt.setFontWeight(weight)
    return fmt


class PythonSyntaxHighlighter(qt.QSyntaxHighlighter):
    def __init__(self, parent=None):
        self.keywordFormat = text_format(qt.Qt.blue, qt.QFont.Bold)
        self.stringFormat = text_format(qt.Qt.darkGreen)
        self.defFormat = text_format(qt.Qt.black, qt.QFont.Bold)
        self.commentFormat = text_format(qt.Qt.lightGray)
        self.decoratorFormat = text_format(qt.Qt.darkGray)

        self.keywords = list(keyword.kwlist)

        self.rules = [
            (qt.QRegExp(r"\b%s\b" % kwd), self.keywordFormat) for kwd in self.keywords
        ] + [
            (qt.QRegExp(r"\bdef\s+([A-Za-z_]+[A-Za-z0-9_]+)\s*\("), self.defFormat),
            (qt.QRegExp(r"\bclass\s+([A-Za-z_]+[A-Za-z0-9_]+)\s*\("), self.defFormat),
            (qt.QRegExp(r"'.*'"), self.stringFormat),
            (qt.QRegExp(r'".*"'), self.stringFormat),
            (qt.QRegExp(r"#.*"), self.commentFormat),
            (qt.QRegExp(r"@[A-Za-z_]+[A-Za-z0-9_]+"), self.decoratorFormat),
        ]

        self.multilineStart = qt.QRegExp(r"(''')|" + r'(""")')
        self.multilineEnd = qt.QRegExp(r"(''')|" + r'(""")')

        super().__init__(parent)

    def highlightBlock(self, text):
        for pattern, format in self.rules:
            exp = qt.QRegExp(pattern)
            index = exp.indexIn(text)
            while index >= 0:
                length = exp.matchedLength()
                if exp.captureCount() > 0:
                    self.setFormat(exp.pos(1), len(str(exp.cap(1))), format)
                else:
                    self.setFormat(exp.pos(0), len(str(exp.cap(0))), format)
                index = exp.indexIn(text, index + length)

        # Multi line strings
        start = self.multilineStart
        end = self.multilineEnd

        self.setCurrentBlockState(0)
        startIndex, skip = 0, 0
        if self.previousBlockState() != 1:
            startIndex, skip = start.indexIn(text), 3
        while startIndex >= 0:
            endIndex = end.indexIn(text, startIndex + skip)
            if endIndex == -1:
                self.setCurrentBlockState(1)
                commentLen = len(text) - startIndex
            else:
                commentLen = endIndex - startIndex + 3
            self.setFormat(startIndex, commentLen, self.stringFormat)
            startIndex, skip = (start.indexIn(text, startIndex + commentLen + 3), 3)


class PythonScriptEditor(qt.QPlainTextEdit):
    INDENT = 4

    def __init__(self, widget):
        super().__init__()
        self.widget = widget

    def lastLine(self):
        text = str(self.toPlainText())
        pos = self.textCursor().position()
        index = text.rfind("\n", 0, pos)
        text = text[index:pos].lstrip("\n")
        return text

    def keyPressEvent(self, event):
        if event.key() == qt.Qt.Key_Return:
            if event.modifiers() & (
                qt.Qt.ShiftModifier | qt.Qt.ControlModifier | qt.Qt.MetaModifier
            ):
                self.widget.commit()
                return
            text = self.lastLine()
            indent = len(text) - len(text.lstrip())
            if text.strip() == "pass" or text.strip().startswith("return "):
                indent = max(0, indent - self.INDENT)
            elif text.strip().endswith(":"):
                indent += self.INDENT
            super().keyPressEvent(event)
            self.insertPlainText(" " * indent)
        elif event.key() == qt.Qt.Key_Tab:
            self.insertPlainText(" " * self.INDENT)
        elif event.key() == qt.Qt.Key_Backspace:
            text = self.lastLine()
            if text and not text.strip():
                cursor = self.textCursor()
                for _ in range(min(self.INDENT, len(text))):
                    cursor.deletePreviousChar()
            else:
                super().keyPressEvent(event)

        else:
            super().keyPressEvent(event)


class PythonConsole(qt.QPlainTextEdit, code.InteractiveConsole):
    def __init__(self, locals=None, parent=None):
        qt.QPlainTextEdit.__init__(self, parent)
        code.InteractiveConsole.__init__(self, locals)
        self.history, self.historyInd = [""], 0
        self.loop = self.interact()
        next(self.loop)

    def setLocals(self, locals):
        self.locals = locals

    def updateLocals(self, locals):
        self.locals.update(locals)

    def interact(self, banner=None):
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        cprt = (
            'Type "help", "copyright", "credits" or "license" ' "for more information."
        )
        if banner is None:
            self.write(
                "Python %s on %s\n%s\n(%s)\n"
                % (sys.version, sys.platform, cprt, self.__class__.__name__)
            )
        else:
            self.write("%s\n" % str(banner))
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                self.new_prompt(prompt)
                yield
                try:
                    line = self.raw_input(prompt)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                more = 0

    def raw_input(self, prompt):
        input = str(self.document().lastBlock().previous().text())
        return input[len(prompt) :]

    def new_prompt(self, prompt):
        self.write(prompt)
        self.newPromptPos = self.textCursor().position()
        self.repaint()

    def write(self, data):
        cursor = qt.QTextCursor(self.document())
        cursor.movePosition(qt.QTextCursor.End, qt.QTextCursor.MoveAnchor)
        cursor.insertText(data)
        self.setTextCursor(cursor)
        self.ensureCursorVisible()

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def flush(self):
        pass

    def push(self, line):
        if self.history[0] != line:
            self.history.insert(0, line)
        self.historyInd = 0

        # prevent console errors to trigger error reporting & patch stdout, stderr
        with patch("sys.excepthook", sys.__excepthook__), patch(
            "sys.stdout", self
        ), patch("sys.stderr", self):
            return code.InteractiveConsole.push(self, line)

    def setLine(self, line):
        cursor = qt.QTextCursor(self.document())
        cursor.movePosition(qt.QTextCursor.End)
        cursor.setPosition(self.newPromptPos, qt.QTextCursor.KeepAnchor)
        cursor.removeSelectedText()
        cursor.insertText(line)
        self.setTextCursor(cursor)

    def keyPressEvent(self, event):
        if event.key() == qt.Qt.Key_Return:
            self.write("\n")
            next(self.loop)
        elif event.key() == qt.Qt.Key_Up:
            self.historyUp()
        elif event.key() == qt.Qt.Key_Down:
            self.historyDown()
        elif event.key() == qt.Qt.Key_Tab:
            self.complete()
        elif event.key() in [qt.Qt.Key_Left, qt.Qt.Key_Backspace]:
            if self.textCursor().position() > self.newPromptPos:
                qt.QPlainTextEdit.keyPressEvent(self, event)
        else:
            qt.QPlainTextEdit.keyPressEvent(self, event)

    def historyUp(self):
        self.setLine(self.history[self.historyInd])
        self.historyInd = min(self.historyInd + 1, len(self.history) - 1)

    def historyDown(self):
        self.setLine(self.history[self.historyInd])
        self.historyInd = max(self.historyInd - 1, 0)

    def complete(self):
        pass

    def _moveCursorToInputLine(self):
        """
        Move the cursor to the input line if not already there. If the cursor
        if already in the input line (at position greater or equal to
        `newPromptPos`) it is left unchanged, otherwise it is moved at the
        end.

        """
        cursor = self.textCursor()
        pos = cursor.position()
        if pos < self.newPromptPos:
            cursor.movePosition(qt.QTextCursor.End)
            self.setTextCursor(cursor)

    def pasteCode(self, source):
        """
        Paste source code into the console.
        """
        self._moveCursorToInputLine()

        for line in interleave(source.splitlines(), itertools.repeat("\n")):
            if line != "\n":
                self.insertPlainText(line)
            else:
                self.write("\n")
                next(self.loop)

    def insertFromMimeData(self, source):
        """
        Reimplemented from QPlainTextEdit.insertFromMimeData.
        """
        if source.hasText():
            self.pasteCode(str(source.text()))
            return


def interleave(seq1, seq2):
    """
    Interleave elements of `seq2` between consecutive elements of `seq1`.

        >>> list(interleave([1, 3, 5], [2, 4]))
        [1, 2, 3, 4, 5]

    """
    iterator1, iterator2 = iter(seq1), iter(seq2)
    leading = next(iterator1)
    for element in iterator1:
        yield leading
        yield next(iterator2)
        leading = element

    yield leading


class Script(object):
    Modified = 1
    MissingFromFilesystem = 2

    def __init__(self, name, script, flags=0, filename=None):
        self.name = name
        self.script = script
        self.flags = flags
        self.filename = filename


class ScriptItemDelegate(qt.QStyledItemDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def displayText(self, script, locale):
        if script.flags & Script.Modified:
            return "*" + script.name
        else:
            return script.name

    def paint(self, painter, option, index):
        script = index.data(qt.Qt.DisplayRole)

        if script.flags & Script.Modified:
            option = qt.QStyleOptionViewItem(option)
            option.palette.setColor(qt.QPalette.Text, qt.QColor(qt.Qt.red))
            option.palette.setColor(qt.QPalette.Highlight, qt.QColor(qt.Qt.darkRed))
        super().paint(painter, option, index)

    def createEditor(self, parent, option, index):
        return qt.QLineEdit(parent)

    def setEditorData(self, editor, index):
        script = index.data(qt.Qt.DisplayRole)
        editor.setText(script.name)

    def setModelData(self, editor, model, index):
        model[index.row()].name = str(editor.text())


def select_row(view, row):
    """
    Select a `row` in an item view
    """
    selmodel = view.selectionModel()
    selmodel.select(view.model().index(row, 0), qt.QItemSelectionModel.ClearAndSelect)


class PrepareSavingSettingsHandler(SettingsHandler):
    """Calls storeSpecificSettings, which is currently not called from non-context handlers."""

    def pack_data(self, widget):
        widget.storeSpecificSettings()
        return super().pack_data(widget)


class OWPythonScript(widget.OWBaseWidget, openclass=True):
    name = "Python Script"
    description = "Write a Python script and run it on input data or models."
    icon = "icons/PythonScript.svg"
    priority = 3150
    keywords = ["file", "program"]

    class Inputs:
        data = Input(
            "data", TomwerScanBase, replaces=["in_data"], default=True, multiple=True
        )
        volume = Input(
            "volume",
            TomwerVolumeBase,
            replaces=["in_volume"],
            default=True,
            multiple=True,
        )
        future_tomo_obj = Input(
            "future_tomo_obj",
            FutureTomwerObject,
            replaces=["in_future_tomo_obj"],
            default=True,
            multiple=True,
        )
        cluster_config = Input(
            "cluster_config",
            SlurmClusterConfiguration,
            replaces=["in_cluster_config"],
            default=True,
            multiple=True,
        )
        object = Input(
            "Object",
            object,
            replaces=["in_object"],
            default=False,
            multiple=True,
            auto_summary=False,
        )
        configuration = Input(
            "configuration",
            TomwerScanBase,
            replaces=["in_configuration"],
            default=True,
            multiple=True,
        )
        serie = Input(
            "serie",
            Serie,
            replaces=["in_serie"],
            default=True,
            multiple=True,
        )

    class Outputs:
        data = Output("data", TomwerScanBase, replaces=["out_data"])
        volume = Output("volume", TomwerVolumeBase, replaces=["out_volume"])
        object = Output("Object", object, replaces=["out_object"], auto_summary=False)
        configuration = Output("configuration", dict, replaces=["out_configuration"])
        future_tomo_obj = Output(
            "future_tomo_obj", FutureTomwerObject, replaces=["out_future_tomo_obj"]
        )
        cluster_config = Output(
            "cluster_config", SlurmClusterConfiguration, replaces=["out_cluster_config"]
        )
        serie = Output("serie", dict, replaces=["out_serie"])

    signal_names = (
        "data",
        "volume",
        "future_tomo_obj",
        "object",
        "configuration",
        "cluster_config",
        "serie",
    )

    settingsHandler = PrepareSavingSettingsHandler()

    libraryListSource = Setting([Script("Hello world", "print('Hello world')\n")])
    currentScriptIndex = Setting(0)
    scriptText = Setting(None, schema_only=True)
    splitterState = Setting(None)

    ewokstaskclass = tomwer.core.process.script.python.PythonScript

    class Error(OWBaseWidget.Error):
        pass

    def __init__(self):
        super().__init__()

        for name in self.signal_names:
            setattr(self, name, {})

        for s in self.libraryListSource:  # pylint: disable=E1133
            s.flags = 0

        self._cachedDocuments = {}

        self.infoBox = gui.vBox(self.controlArea, "Info")
        gui.label(
            self.infoBox,
            self,
            "<p>Execute python script.</p><p>Input variables:<ul><li> "
            + "<li>".join(map("in_{0}, in_{0}s".format, self.signal_names))
            + "</ul></p><p>Output variables:<ul><li>"
            + "<li>".join(map("out_{0}".format, self.signal_names))
            + "</ul></p>",
        )

        self.libraryList = itemmodels.PyListModel(
            [],
            self,
            flags=qt.Qt.ItemIsSelectable | qt.Qt.ItemIsEnabled | qt.Qt.ItemIsEditable,
        )

        self.libraryList.wrap(self.libraryListSource)

        self.controlBox = gui.vBox(self.controlArea, "Library")
        self.controlBox.layout().setSpacing(1)

        self.libraryView = qt.QListView(
            editTriggers=qt.QListView.DoubleClicked | qt.QListView.EditKeyPressed,
            sizePolicy=qt.QSizePolicy(qt.QSizePolicy.Ignored, qt.QSizePolicy.Preferred),
        )
        self.libraryView.setItemDelegate(ScriptItemDelegate(self))
        self.libraryView.setModel(self.libraryList)

        self.libraryView.selectionModel().selectionChanged.connect(
            self.onSelectedScriptChanged
        )
        self.controlBox.layout().addWidget(self.libraryView)

        w = itemmodels.ModelActionsWidget()

        self.addNewScriptAction = action = qt.QAction("+", self)
        action.setToolTip("Add a new script to the library")
        action.triggered.connect(self.onAddScript)
        w.addAction(action)

        action = qt.QAction(unicodedata.lookup("MINUS SIGN"), self)
        action.setToolTip("Remove script from library")
        action.triggered.connect(self.onRemoveScript)
        w.addAction(action)

        action = qt.QAction("Update", self)
        action.setToolTip("Save changes in the editor to library")
        action.setShortcut(qt.QKeySequence(qt.QKeySequence.Save))
        action.triggered.connect(self.commitChangesToLibrary)
        w.addAction(action)

        action = qt.QAction("More", self, toolTip="More actions")

        new_from_file = qt.QAction("Import Script from File", self)
        save_to_file = qt.QAction("Save Selected Script to File", self)
        restore_saved = qt.QAction("Undo Changes to Selected Script", self)
        save_to_file.setShortcut(qt.QKeySequence(qt.QKeySequence.SaveAs))

        new_from_file.triggered.connect(self.onAddScriptFromFile)
        save_to_file.triggered.connect(self.saveScript)
        restore_saved.triggered.connect(self.restoreSaved)

        menu = qt.QMenu(w)
        menu.addAction(new_from_file)
        menu.addAction(save_to_file)
        menu.addAction(restore_saved)
        action.setMenu(menu)
        button = w.addAction(action)
        button.setPopupMode(qt.QToolButton.InstantPopup)

        w.layout().setSpacing(1)

        self.controlBox.layout().addWidget(w)

        self.execute_button = gui.button(
            self.controlArea, self, "Run", callback=self.commit
        )

        self.splitCanvas = qt.QSplitter(qt.Qt.Vertical, self.mainArea)
        self.mainArea.layout().addWidget(self.splitCanvas)

        self.defaultFont = defaultFont = (
            "Monaco" if sys.platform == "darwin" else "Courier"
        )

        self.textBox = gui.vBox(self, "Python Script")
        self.splitCanvas.addWidget(self.textBox)
        self.text = PythonScriptEditor(self)
        self.textBox.layout().addWidget(self.text)

        self.textBox.setAlignment(qt.Qt.AlignVCenter)
        self.text.setTabStopWidth(4)

        self.text.modificationChanged[bool].connect(self.onModificationChanged)

        self.saveAction = action = qt.QAction("&Save", self.text)
        action.setToolTip("Save script to file")
        action.setShortcut(qt.QKeySequence(qt.QKeySequence.Save))
        action.setShortcutContext(qt.Qt.WidgetWithChildrenShortcut)
        action.triggered.connect(self.saveScript)

        self.consoleBox = gui.vBox(self, "Console")
        self.splitCanvas.addWidget(self.consoleBox)
        self.console = PythonConsole({}, self)
        self.consoleBox.layout().addWidget(self.console)
        self.console.document().setDefaultFont(qt.QFont(defaultFont))
        self.consoleBox.setAlignment(qt.Qt.AlignBottom)
        self.console.setTabStopWidth(4)

        select_row(self.libraryView, self.currentScriptIndex)

        self.restoreScriptText()

        self.splitCanvas.setSizes([2, 1])
        if self.splitterState is not None:
            self.splitCanvas.restoreState(qt.QByteArray(self.splitterState))

        self.splitCanvas.splitterMoved[int, int].connect(self.onSpliterMoved)
        self.controlArea.layout().addStretch(1)
        self.resize(800, 600)

    def storeSpecificSettings(self):
        self.saveScriptText()

    def restoreScriptText(self):
        if self.scriptText is not None:
            current = self.text.toPlainText()
            # do not mark scripts as modified
            if self.scriptText != current:
                self.text.document().setPlainText(self.scriptText)

    def saveScriptText(self):
        self.scriptText = self.text.toPlainText()

    def handle_input(self, obj, id, signal):
        id = id[0]
        dic = getattr(self, signal)
        if obj is None:
            if id in dic.keys():
                del dic[id]
        else:
            dic[id] = obj

    @Inputs.data  # noqa F811
    def set_data(self, data, id):
        self.handle_input(data, id, "data")

    @Inputs.volume  # noqa F811
    def set_volume(self, volume, id):
        self.handle_input(volume, id, "volume")

    @Inputs.object  # noqa F811
    def set_object(self, data, id):  # noqa F811
        self.handle_input(data, id, "object")

    @Inputs.configuration  # noqa F811
    def set_configuration(self, data, id):
        self.handle_input(data, id, "configuration")

    @Inputs.future_tomo_obj  # noqa F811
    def set_future_tomo_object(self, data, id):  # noqa F811 pylint: disable=E0102
        self.handle_input(data, id, "future_tomo_obj")

    @Inputs.cluster_config  # noqa F811
    def set_cluster_config(self, data, id):  # noqa F811 pylint: disable=E0102
        self.handle_input(data, id, "cluster_config")

    @Inputs.serie  # noqa F811
    def set_serie(self, data, id):
        self.handle_input(data, id, "serie")

    def handleNewSignals(self):
        self.commit()

    def selectedScriptIndex(self):
        rows = self.libraryView.selectionModel().selectedRows()
        if rows:
            return [i.row() for i in rows][0]
        else:
            return None

    def setSelectedScript(self, index):
        select_row(self.libraryView, index)

    def onAddScript(self, *args):
        self.libraryList.append(Script("New script", self.text.toPlainText(), 0))
        self.setSelectedScript(len(self.libraryList) - 1)

    def onAddScriptFromFile(self, *args):  # pragma: no cover
        filename, _ = qt.QFileDialog.getOpenFileName(
            self,
            "Open Python Script",
            os.path.expanduser("~/"),
            "Python files (*.py)\nAll files(*.*)",
        )
        if filename:
            name = os.path.basename(filename)
            # TODO: use `tokenize.detect_encoding`
            with open(filename, encoding="utf-8") as f:
                contents = f.read()
            self.libraryList.append(Script(name, contents, 0, filename))
            self.setSelectedScript(len(self.libraryList) - 1)

    def onRemoveScript(self, *args):
        index = self.selectedScriptIndex()
        if index is not None:
            del self.libraryList[index]
            select_row(self.libraryView, max(index - 1, 0))

    def onSaveScriptToFile(self, *args):
        index = self.selectedScriptIndex()
        if index is not None:
            self.saveScript()

    def onSelectedScriptChanged(self, selected, deselected):
        index = [i.row() for i in selected.indexes()]
        if index:
            current = index[0]
            if current >= len(self.libraryList):
                self.addNewScriptAction.trigger()
                return

            self.text.setDocument(self.documentForScript(current))
            self.currentScriptIndex = current

    def documentForScript(self, script=0):
        if type(script) != Script:
            script = self.libraryList[script]
        if script not in self._cachedDocuments:
            doc = qt.QTextDocument(self)
            doc.setDocumentLayout(qt.QPlainTextDocumentLayout(doc))
            doc.setPlainText(script.script)
            doc.setDefaultFont(qt.QFont(self.defaultFont))
            doc.highlighter = PythonSyntaxHighlighter(doc)
            doc.modificationChanged[bool].connect(self.onModificationChanged)
            doc.setModified(False)
            self._cachedDocuments[script] = doc
        return self._cachedDocuments[script]

    def commitChangesToLibrary(self, *args):
        index = self.selectedScriptIndex()
        if index is not None:
            self.libraryList[index].script = self.text.toPlainText()
            self.text.document().setModified(False)
            self.libraryList.emitDataChanged(index)

    def onModificationChanged(self, modified):
        index = self.selectedScriptIndex()
        if index is not None:
            self.libraryList[index].flags = Script.Modified if modified else 0
            self.libraryList.emitDataChanged(index)

    def onSpliterMoved(self, pos, ind):
        self.splitterState = bytes(self.splitCanvas.saveState())

    def restoreSaved(self):
        index = self.selectedScriptIndex()
        if index is not None:
            self.text.document().setPlainText(self.libraryList[index].script)
            self.text.document().setModified(False)

    def saveScript(self):  # pragma: no cover
        index = self.selectedScriptIndex()
        if index is not None:
            script = self.libraryList[index]
            filename = script.filename
        else:
            filename = os.path.expanduser("~/")

        filename, _ = qt.QFileDialog.getSaveFileName(
            self, "Save Python Script", filename, "Python files (*.py)\nAll files(*.*)"
        )

        if filename:
            fn = ""
            head, tail = os.path.splitext(filename)
            if not tail:
                fn = head + ".py"
            else:
                fn = filename

            f = open(fn, "w")
            f.write(self.text.toPlainText())
            f.close()

    def initial_locals_state(self):
        d = {}
        for name in self.signal_names:
            value = getattr(self, name)
            all_values = list(value.values())
            one_value = all_values[0] if len(all_values) == 1 else None
            d["in_" + name + "s"] = all_values
            d["in_" + name] = one_value
        return d

    def commit(self):
        self.Error.clear()
        self._script = str(self.text.toPlainText())
        lcls = self.initial_locals_state()
        lcls["_script"] = str(self.text.toPlainText())
        self.console.updateLocals(lcls)
        self.console.write("\nRunning script:\n")
        self.console.push("exec(_script)")
        self.console.new_prompt(sys.ps1)
        for signal in self.signal_names:
            out_var = self.console.locals.get("out_" + signal)
            signal_type = getattr(self.Outputs, signal).type
            if not isinstance(out_var, signal_type) and out_var is not None:
                self.Error.add_message(
                    signal,
                    "'{}' has to be an instance of '{}'.".format(
                        signal, signal_type.__name__
                    ),
                )
                getattr(self.Error, signal)()
                out_var = None
            getattr(self.Outputs, signal).send(out_var)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/06/2018"


import subprocess
from silx.gui import qt
import tomwer.core.utils.gpu
from tomwer.core.process.reconstruction.lamino.tofu import has_tofu
import logging
from .TofuOptionLoader import _TofuOptionLoader, _getterSetter

_logger = logging.getLogger(__name__)


class ExpertTofuWidget(_TofuOptionLoader, qt.QWidget):
    """
    Widget used to define the parameters for the tofu script
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent)
        self.setLayout(qt.QVBoxLayout())

        self._verbose = qt.QCheckBox("verbose", parent=self)
        self._verbose.setChecked(True)
        self.layout().addWidget(self._verbose)

        # retries group
        self._retriesGB = qt.QGroupBox("retries parameters", parent=self)
        self._retriesGB.setLayout(qt.QGridLayout())
        self._retriesGB.layout().addWidget(
            qt.QLabel("n retries:", parent=self._retriesGB), 0, 0
        )
        self._retriesSP = qt.QSpinBox(parent=self._retriesGB)
        self._retriesSP.setMinimum(0)
        self._retriesGB.layout().addWidget(self._retriesSP, 0, 1)
        self._retriesGB.layout().addWidget(
            qt.QLabel("timeout:", parent=self._retriesGB), 1, 0
        )
        self._timeoutLE = qt.QLineEdit("0", parent=self._retriesGB)
        validator = qt.QDoubleValidator(parent=self._timeoutLE)
        validator.setBottom(0.0)
        self._timeoutLE.setValidator(validator)
        self._retriesGB.layout().addWidget(self._timeoutLE, 1, 1)
        self.layout().addWidget(self._retriesGB)

        # GPU group
        self._gpuGrp = GpuGroupBox(parent=self)
        self.layout().addWidget(self._gpuGrp)

        # options text edit
        self._qteOptions = OptionEdit(parent=self)
        self.layout().addWidget(self._qteOptions)

        # expose API
        self.getAdditionalRecoOptions = self._qteOptions.getAdditionalRecoOptions
        self.addAdditionalRecoOption = self._qteOptions.addAdditionalRecoOption
        self.setAdditionalRecoOptions = self._qteOptions.setAdditionalRecoOptions
        self.getAdditionalPreprocessOptions = (
            self._qteOptions.getAdditionalPreprocessOptions
        )
        self.addAdditionalPreprocessOption = (
            self._qteOptions.addAdditionalPreprocessOption
        )
        self.setAdditionalPreprocessOptions = (
            self._qteOptions.setAdditionalPreprocessOptions
        )
        self.getHighLimit = self._gpuGrp.getHighLimit

        options = {
            "verbose": _getterSetter(
                getter=self.requireVerbose, setter=self.setVerbose
            ),
            "retries": _getterSetter(getter=self.getNRetries, setter=self.setNRetries),
            "retry-timeout": _getterSetter(
                getter=self.getNRetries, setter=self.setRetriesTimeout
            ),
            "slices-per-device": _getterSetter(
                getter=self.getSlicePerDevice, setter=self.setSlicePerDevice
            ),
            "slice-memory-coeff": _getterSetter(
                getter=self.getSliceMemoryCoeff, setter=self.setSliceMemoryCoeff
            ),
        }
        _TofuOptionLoader.__init__(self, options=options)

    def getNRetries(self):
        """

        :return: return number of retry before avoiding the reconstruction
        """
        return self._retriesSP.value()

    def setNRetries(self, value):
        """ """
        try:
            self._retriesSP.setValue(int(value))
        except Exception as error:
            _logger.warning(f"Fails to set N tries, error is {error}")

    def getRetriesTimeout(self):
        """

        :return: return time (in second) of the time out for the tofu
                 reconstruction
        """
        return float(self._timeoutLE.text())

    def setRetriesTimeout(self, value):
        """ """
        if value is None:
            return
        try:
            self._timeoutLE.setText(str(value))
        except Exception as error:
            _logger.warning(f"Fails to set N tries, error is {error}")

    def getSliceMemoryCoeff(self):
        """

        :return: None if want default, otherwise value between 0.01 and 0.9
        """
        return self._gpuGrp.getSliceMemoryCoeff()

    def getSlicePerDevice(self):
        """

        :return: None if we don't want to activate this option. Otherwise int
                 from 1 to inf
        """
        return self._gpuGrp.getSlicePerDevice()

    def setSlicePerDevice(self, nslice):
        if nslice is not None:
            self._gpuGrp.setSlicePerDevice(nslice)

    def setSliceMemoryCoeff(self, value):
        if value is not None:
            self._gpuGrp.setSliceMemoryCoeff(value)

    def requireVerbose(self):
        """

        :return: True if the user require verbose for reconstruction
        """
        return self._verbose.isChecked()

    def setVerbose(self, value=True):
        _value = value
        if _value == "":
            _value = True
        self._verbose.setChecked(_value)

    def loadFromScan(self, scanID):
        # nothing to be done here
        pass

    def resetAdditionalRecoOptions(self):
        self._qteOptions.resetAdditionalRecoOptions()

    def resetAdditionalPreprocessOptions(self):
        self._qteOptions.resetAdditionalPreprocessOptions()


class OptionEdit(qt.QWidget):
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(qt.QLabel("options:"))
        self._tabWidget = qt.QTabWidget(parent=self)

        # recons opt
        self._qteAddReconsOptions = qt.QTextEdit(parent=self._tabWidget)
        self._addAdditionalReconsOptions()
        self._tabWidget.addTab(self._qteAddReconsOptions, "Additional reco options")
        self._addAdditionalReconsOptions()
        # ffc options
        self._qteAddPreprocessOptions = qt.QTextEdit(parent=self._tabWidget)
        self._addAdditionalReconsOptions()
        self._tabWidget.addTab(
            self._qteAddPreprocessOptions, "Additional preprocess options"
        )
        self._addAdditionalPreprocessOptions()
        # fixed options
        self._qteFixedOptions = qt.QTextEdit(parent=self._tabWidget)
        self._qteFixedOptions.setDisabled(True)
        self._addFixedOptions()
        self._tabWidget.addTab(self._qteFixedOptions, "Options managed by the GUI")
        self.layout().addWidget(self._tabWidget)
        # reco help
        self._recoHelpScrollArea = qt.QScrollArea(parent=self)
        self._qteTofuRecoHelp = qt.QLabel("", parent=self._recoHelpScrollArea)
        self._addTofuRecoHelp()
        self._recoHelpScrollArea.setWidget(self._qteTofuRecoHelp)
        self._tabWidget.addTab(self._recoHelpScrollArea, "See tofu reco help")
        # ffc help
        self._preprocessHelpScrollArea = qt.QScrollArea(parent=self)
        self._qteTofuPreprocessHelp = qt.QLabel(
            "", parent=self._preprocessHelpScrollArea
        )
        self._addTofuPreprocessHelp()
        self._preprocessHelpScrollArea.setWidget(self._qteTofuPreprocessHelp)
        self._tabWidget.addTab(
            self._preprocessHelpScrollArea, "See tofu preprocess help"
        )

        # add some tooltip
        self._qteAddReconsOptions.setToolTip(
            "If you want some options to be "
            "passed to tofu reco which are not "
            "present in 'options managed' "
            "then add the option and value if "
            "necessary in the list"
        )
        self._qteAddPreprocessOptions.setToolTip(
            "If you want some options to be "
            "passed to tofu preprocess which are not "
            "present in 'options managed' "
            "then add the option and value if "
            "necessary in the list"
        )
        self._qteFixedOptions.setToolTip(
            "This list all the options "
            "that are generated automatically "
            "with there values (if necessary) "
            "by the interface"
        )
        self._qteTofuRecoHelp.setToolTip("See help from the used tofu version")

    def _addFixedOptions(self):
        """
        Add information about the fixed option (the one always generated
        in the tofu reco call)
        :return:
        """
        options = [
            "z-parameter []",
            "retry-timeout []",
            "retries []",
            "verbose",
            "slices-per-device []",
            "slice-memory-coeff []",
            "projections []",
            "darks []",
            "flats []",
            "flats2 []",
            "fix-nan-and-inf",
            "number []",
            "pixel-size []",
            "volume-angle-x []",
            "volume-angle-y []",
            "volume-angle-z []",
            "region []",
            "output []",
            "dry-run (activated if no output specify)",
            "center-position-x []",
            "center-position-z []",
            "overall-angle []",
            "axis-angle-x []",
            "axis-angle-y []",
            "x-region []",
            "y-region []",
            "regularization-rate",
            "absorptivity",
            "thresholding-rate",
            "retrieval-method",
            "energy",
            "propagation-distance",
        ]
        self._qteFixedOptions.setText(" --".join(options))

    def _addAdditionalReconsOptions(self):
        self._qteAddReconsOptions.setText(
            " ".join(
                (
                    "--resize 0",
                    "--projection-padding-mode clamp_to_edge",
                    "--genreco-padding-mode clamp_to_edge",
                )
            )
        )

    def _addAdditionalPreprocessOptions(self):
        self._qteAddPreprocessOptions.setText(
            " ".join(
                (
                    "--resize 0",
                    "--projection-padding-mode clamp_to_edge",
                    "--fix-nan-and-inf",
                    "--projection-filter none",
                )
            )
        )

    def addAdditionalRecoOption(self, option, value):
        self._qteAddReconsOptions.setText(
            " ".join(
                (
                    self._qteAddReconsOptions.toPlainText(),
                    "--" + option + " " + str(value),
                )
            )
        )

    def addAdditionalPreprocessOption(self, option, value):
        self._qteAddPreprocessOptions.setText(
            " ".join(
                (
                    self._qteAddPreprocessOptions.toPlainText(),
                    "--" + option + " " + str(value),
                )
            )
        )

    def setAdditionalRecoOptions(self, text):
        self._qteAddReconsOptions.setText(text)

    def setAdditionalPreprocessOptions(self, text):
        self._qteAddPreprocessOptions.setText(text)

    def getAdditionalRecoOptions(self):
        return self._qteAddReconsOptions.toPlainText().replace("\n", " ")

    def getAdditionalPreprocessOptions(self):
        return self._qteAddPreprocessOptions.toPlainText().replace("\n", " ")

    def resetAdditionalRecoOptions(self):
        self._qteAddReconsOptions.clear()

    def resetAdditionalPreprocessOptions(self):
        self._qteAddPreprocessOptions.clear()

    def _getNodeParameters(self):
        """Add the value of the detain options + some extra parameters to be
        passed as 'free string'

        :return:
        :rtype: tuple (dict, list)
        """
        _ddict = {}
        for option in self._options:
            _ddict[option] = self._options[option].getter()
        return _ddict, self.getAdditionalRecoOptions()

    def _addTofuRecoHelp(self):
        if has_tofu() is True:
            try:
                process = subprocess.Popen(
                    ["tofu", "reco", "--help"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                stdout, stderr = process.communicate()
                if (len(stdout.decode("utf-8"))) > 0:
                    self._qteTofuRecoHelp.setText(stdout.decode("utf-8"))
                else:
                    self._qteTofuRecoHelp.setText(stderr.decode("utf-8"))
            except Exception:
                try:
                    process.kill()
                    process.wait()
                except Exception:
                    self._qteTofuRecoHelp.setText("Tofu reco help not found")
        else:
            self._qteTofuRecoHelp.setText("Tofu not found")
        self._qteTofuRecoHelp.setTextInteractionFlags(qt.Qt.TextSelectableByMouse)
        self._qteTofuRecoHelp.setAlignment(qt.Qt.AlignTop)

    def _addTofuPreprocessHelp(self):
        if has_tofu() is True:
            try:
                process = subprocess.Popen(
                    ["tofu", "preprocess", "--help"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                stdout, stderr = process.communicate()
                if (len(stdout.decode("utf-8"))) > 0:
                    self._qteTofuPreprocessHelp.setText(stdout.decode("utf-8"))
                else:
                    self._qteTofuPreprocessHelp.setText(stderr.decode("utf-8"))
            except Exception:
                try:
                    process.kill()
                    process.wait()
                except Exception:
                    self._qteTofuPreprocessHelp.setText(
                        "Tofu preprocess help not found"
                    )
        else:
            self._qteTofuPreprocessHelp.setText("Tofu not found")
        self._qteTofuPreprocessHelp.setTextInteractionFlags(qt.Qt.TextSelectableByMouse)
        self._qteTofuPreprocessHelp.setAlignment(qt.Qt.AlignTop)


class GpuGroupBox(qt.QGroupBox):
    """
    Class grouping all the options relative to GPU
    """

    gpuOptionChanged = qt.Signal()
    """Signal emitted when the option of the gpu are modified"""

    class SlicePerDeviceWidget(qt.QWidget):
        def __init__(self, parent):
            qt.QWidget.__init__(self, parent)
            self.setLayout(qt.QVBoxLayout())
            self.layout().setContentsMargins(0, 0, 0, 0)
            self._noneLabel = qt.QLabel("None", parent=self)
            self.layout().addWidget(self._noneLabel)
            self._slicePerDevice = qt.QSpinBox(parent=self)
            self._slicePerDevice.setMinimum(1)
            self._slicePerDevice.setMaximum(1000000)
            self.layout().addWidget(self._slicePerDevice)
            self._slicePerDevice.setVisible(False)

        def setValue(self, value):
            self._slicePerDevice.setVisible(value is not None)
            self._noneLabel.setVisible(value is None)

            if value is not None:
                self._slicePerDevice.setValue(value)

        def getValue(self):
            if self._noneLabel.isVisible():
                return None
            else:
                return self._slicePerDevice.value()

        def activeValueMode(self, b):
            self._slicePerDevice.setVisible(b is True)
            self._slicePerDevice.setEnabled(b is True)
            self._noneLabel.setVisible(b is False)
            self.setVisible(True)

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, "GPU optional parameters", parent)
        self.setLayout(qt.QGridLayout())

        self._sliceMemoryCoeffRB = qt.QRadioButton("slice memory coeff", parent=self)

        self.layout().addWidget(self._sliceMemoryCoeffRB, 0, 0)

        self._sliceMemoryCoeff = qt.QLineEdit("0.8", parent=self)
        validator = qt.QDoubleValidator(parent=self._sliceMemoryCoeff)
        validator.setBottom(0.01)
        validator.setTop(0.9)
        self._sliceMemoryCoeff.setValidator(validator)

        self.layout().addWidget(self._sliceMemoryCoeff, 0, 2)
        self._sliceMemoryCoeffRB.toggled.connect(self._mvToMemoCoeff)

        self._slicePerDeviceRB = qt.QRadioButton("slices per device", parent=self)
        self.layout().addWidget(self._slicePerDeviceRB, 1, 0)

        self._slicePerDevice = self.SlicePerDeviceWidget(parent=self)
        self.layout().addWidget(self._slicePerDevice, 1, 2)
        self._slicePerDevice.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )

        self._slicePerDeviceRB.toggled.connect(self._mvToSlicePerDevice)

        self._autoRB = qt.QRadioButton("auto", parent=self)
        self.layout().addWidget(self._autoRB, 2, 0)
        self._autoRB.toggled.connect(self._mvToAuto)
        self._autoRB.setChecked(True)

        for _radioButton in (
            self._slicePerDeviceRB,
            self._sliceMemoryCoeffRB,
            self._autoRB,
        ):
            _radioButton.toggled.connect(self.__gpuOptionChange)
        self._slicePerDevice._slicePerDevice.valueChanged.connect(
            self.__gpuOptionChange
        )
        self._sliceMemoryCoeff.textChanged.connect(self.__gpuOptionChange)

    def _mvToAuto(self):
        """Move to auto mode"""
        self._sliceMemoryCoeff.setText("0.8")
        self._sliceMemoryCoeff.setVisible(True)
        self._sliceMemoryCoeff.setEnabled(False)
        self._slicePerDevice.setValue(None)
        self._slicePerDevice.setVisible(True)
        self._slicePerDevice.setEnabled(False)

    def _mvToMemoCoeff(self):
        self._slicePerDevice.setEnabled(False)
        self._sliceMemoryCoeff.setVisible(True)
        self._sliceMemoryCoeff.setEnabled(True)
        self._slicePerDevice.setValue(None)
        self._slicePerDevice.setVisible(False)

    def _mvToSlicePerDevice(self):
        self._slicePerDevice.setEnabled(True)
        self._sliceMemoryCoeff.setVisible(False)
        self._sliceMemoryCoeff.setEnabled(False)
        self._slicePerDevice.activeValueMode(True)

    def getSliceMemoryCoeff(self):
        if self._sliceMemoryCoeff.isVisible():
            return float(self._sliceMemoryCoeff.text())
        else:
            return None

    def getSlicePerDevice(self):
        if self._slicePerDevice.isVisible():
            return self._slicePerDevice.getValue()
        else:
            return None

    def setSlicePerDevice(self, value):
        self._slicePerDevice.setEnabled(True)
        self._slicePerDevice.setVisible(True)
        try:
            self._slicePerDevice.setValue(int(value))
        except Exception as error:
            _logger.warning(f"Fails to set slice per device, error is {error}")

    def setSliceMemoryCoeff(self, value):
        self._sliceMemoryCoeff.setVisible(True)
        self._sliceMemoryCoeff.setEnabled(True)
        try:
            self._sliceMemoryCoeff.setText(value)
        except Exception as error:
            _logger.warning(f"Fails to set memory coefficient, error is {error}")

    def getHighLimit(self):
        if self._slicePerDeviceRB.isChecked():
            n_gpu = tomwer.core.utils.gpu.getNumberOfDevice() or 1
            return self._slicePerDevice._slicePerDevice.value() * n_gpu
        else:
            return None

    def __gpuOptionChange(self, *args, **kwargs):
        self.gpuOptionChanged.emit()
